/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "events.hpp"
#include "registry.hpp"
#include "logger.hpp"
#include "network.hpp"
#include "nonce.hpp"
#include "utils.hpp"
#include "calls.hpp"
#include "proc.hpp"

#include <strings.hpp>
#include <keyfile.hpp>

#include <thread>
#include <functional>
#include <vector>
#include <unordered_set>

using namespace server;
using namespace std::literals::chrono_literals;
using namespace std::literals::string_literals;

namespace {
// all names we may locally answer for...
std::unordered_set<std::string> localhosts;     // names we answer locally as
std::unordered_set<nonce_t> nonces;             // local size nonces issued
std::string realm = "Coventry", algorithm = "SHA-256,MD5";
time_t max_expires = 300;

// Local functions
auto unquote_view(const char *str)
{
    if(!str)
        str = "";

    std::string_view text(str);
    if(text.size() < 2)
        return text;

    if(text.front() == text.back() && text.front() == '\"') {
        text.remove_prefix(1);
        text.remove_suffix(1);
    }
    return text;
}

auto is_local(osip_uri *uri)
{
    if(!uri || !uri->host)
        return false;

    return std::any_of(localhosts.begin(), localhosts.end(), [uri](std::string_view localhost) {
        return localhost == uri->host;
    });
}

auto is_authorized(const event_t& event, bool ignore_nonce = false)
{
    auto request = event.request();
    if(request) {
        auto authorization = event.authorize();

        if(!authorization || !authorization->realm) {
            debug(5, "no authorization record offered");
            return false;
        }

        debug(5, "has authorization record");
        if(realm != event.realm()) {
            debug(4, "auth: realm {} is not {}", event.realm(), realm);
            return false;
        }

        if(event.algorithm() != "MD5" && event.algorithm() != "SHA-256") {
            debug(4, "auth: unknown digest");
            return false;
        }

        if(authorization->username && authorization->nonce && authorization->response) {
            auto nonce = nonce_t(event.nonce());
            if(ignore_nonce || nonces.count(nonce))
                return true;
        }
    }
    return false;
}

// Sip operations
auto server_options(const event_t& event, std::function<void (exosip::msg_t)> fill) {
    auto context = event.context();
    auto tid = event.tid();
    auto msg = exosip::msg_t(nullptr);

    exosip::guard_t lock(context);
    eXosip_options_build_answer(context, tid, SIP_OK, &msg);
    if(!msg)
        return false;

    fill(msg);
    eXosip_options_send_answer(context, tid, SIP_OK, msg);
    return true;
}

// Challenges
auto challenge_create(osip_message_t *msg)
{
    using namespace std::literals::string_literals;

    auto nonce = nonce_t();
    nonces.emplace(nonce);
    auto challenge = "Digest realm=\"" + realm + "\", nonce=\"" + nonce.hex() + "\",algorithm=\"" + algorithm + "\"";
    debug(5, "challenge {}", challenge);
    osip_message_set_header(msg, WWW_AUTHENTICATE, challenge.c_str());
    osip_message_set_header(msg, "Expires", std::to_string(max_expires).c_str());
}

void challenge_message(const event_t& event)
{
    auto context = event.context();
    auto tid = event.tid();
    auto msg = exosip::msg_t(nullptr);
    auto authorize = event.authorize();

    // if we already have a failed or partial auth record, call it forbidden...
    if(authorize) {
        if(!authorize->realm || realm != event.realm()) {
            debug(4, "auth: realm {} vs {}", realm, event.realm());
            return event.fail_message(SIP_FORBIDDEN);
        }

        if(event.algorithm() != "MD5" && event.algorithm() != "SHA-256") {
            debug(4, "auth: unknown digest {}", event.algorithm());
            return event.fail_message(SIP_FORBIDDEN);
        }
    }

    exosip::guard_t lock(context);
    eXosip_message_build_answer(context, tid, SIP_UNAUTHORIZED, &msg);
    challenge_create(msg);
    eXosip_message_send_answer(context, tid, SIP_UNAUTHORIZED, msg);
}

void challenge_call(const event_t& event)
{
    auto context = event.context();
    auto tid = event.tid();
    auto msg = exosip::msg_t(nullptr);
    auto authorize = event.authorize();

    // if we already have a failed or partial auth record, call it forbidden...
    if(authorize) {
        if(!authorize->realm || realm != event.realm())
            return event.fail_message(SIP_FORBIDDEN);

        if(event.algorithm() != "MD5" && event.algorithm() != "SHA-256") {
            debug(4, "auth: unknown digest {}", event.algorithm());
            return event.fail_message(SIP_FORBIDDEN);
        }
    }

    debug(5, "sending call challenge");
    exosip::guard_t lock(context);
    eXosip_call_build_answer(context, tid, SIP_UNAUTHORIZED, &msg);
    challenge_create(msg);
    eXosip_call_send_answer(context, tid, SIP_UNAUTHORIZED, msg);
}

// Event dispatch
void do_options(const event_t& event) {  // TODO: Registry dispatch...
    auto to = event.to();
    if(!to || !to->url || !to->url->host) {
        debug(3, "incomplete message request");
        event.fail_message(SIP_ADDRESS_INCOMPLETE);
    }
    else if(!is_local(to->url)) {
        server_options(event, [](exosip::msg_t msg){
            allow_remote(msg);
        });
    }
    else if(!to->url->username) {
        server_options(event, [](exosip::msg_t msg){
            allow_local(msg);
        });
    }
    else
        exec::option_user(event);
}

void do_publish(const event_t& event)
{
    auto to = event.to();
    auto body = event.body();
    auto type = event.content();
    auto publish = event.publish();
    auto expires = event.remains() / 1000;
    //auto view = content_view(body);
    auto local = is_local(to->url);

    if(!to || !to->url || !to->url->username || !to->url->host) {
        debug(3, "incomplete message request");
        event.fail_message(SIP_ADDRESS_INCOMPLETE);
    }
    else if(!body || publish.empty() || expires < 10) {
        debug(4, "unacceptable publish input {}", type);
        event.fail_message(SIP_NOT_ACCEPTABLE_HERE);
    }
    else if(!is_authorized(event)) {
        debug(4, "request publish authorization");
        challenge_message(event);
    }
    else if(!registry::is_extension(event.user())) {
        debug(4, "invalid publish extension {}", event.user());
        event.fail_message(SIP_FORBIDDEN);
    }
    else if(!local) {
        log_warn("remote publishing not enabled; to={}", to->url->host);
        event.fail_message(SIP_NOT_FOUND);
    }
    else
        exec::publish_user(event);
}

void do_message(const event_t& event)
{
    auto to = event.to();
    auto body = event.body();
    auto type = event.content();
    auto view = content_view(body);
    auto local = is_local(to->url);

    if(!to || !to->url || !to->url->username || !to->url->host) {
        debug(3, "incomplete message request");
        event.fail_message(SIP_ADDRESS_INCOMPLETE);
    }
    else if(!body || type != "text/plain") {
        // TODO: some types we just ignore for now...
        if(type == "message/imdn+xml") {
            debug(4, "ignored message input {}", type);
            event.fail_message(SIP_OK);
        }
        else {
            debug(4, "unacceptable message input {}", type);
            event.fail_message(SIP_NOT_ACCEPTABLE_HERE);
        }
    }
    else if(view.size() > 160) {
        debug(4, "message too large");
        event.fail_message(SIP_MESSAGE_TOO_LARGE);
    }
    else if(!util::u8verify(view)) {
        debug(4, "invalid utf8 message input");
        event.fail_message(SIP_NOT_ACCEPTABLE_HERE);
    }
    else if(!is_authorized(event)) {
        debug(4, "request message authorization");
        challenge_message(event);
    }
    else if(!registry::is_extension(event.user())) {
        debug(4, "invalid message extension {}", event.user());
        event.fail_message(SIP_FORBIDDEN);
    }
    else if(local)
        exec::message_user(event);
    else {
        log_warn("remote messaging not enabled; to={}", to->url->host);
        event.fail_message(SIP_NOT_FOUND);
    }
}

void do_invite(const event_t& event)
{
    auto to = event.to();
    auto body = event.body();
    auto type = event.content();
    auto local = is_local(to->url);

    if(!to || !to->url || !to->url->username || !to->url->host) {
        debug(3, "incomplete call request");
        event.fail_call(SIP_ADDRESS_INCOMPLETE);
    }
    else if(!body || type != "application/sdp") {
        debug(4, "unacceptable call input {}", type);
        event.fail_call(SIP_NOT_ACCEPTABLE_HERE);
    }
    else if(!is_authorized(event)) {
        debug(4, "request call authorization");
        challenge_call(event);
    }
    else if(!registry::is_extension(event.user())) {
        debug(4, "invalid call extension {}", event.user());
        event.fail_call(SIP_FORBIDDEN);
    }
    else if(local)
        exec::invite_user(event);
    else {
        log_warn("remote calling not enabled; to={}", to->url->host);
        event.fail_call(SIP_NOT_FOUND);
    }
}

void do_bye(const event_t& event)
{
    auto cid = event.cid();
    debug(4, "bye for call {}", cid);
    //event.fail_call(SIP_NOT_FOUND);
}

void do_register(const event_t& event)
{
    // If we have register without expiration, we ignore the nonce cache and
    // later match the registry nonce.  Some clients terminate by setting a
    // registration expiration of "0" with a pre-stuffed auth header, and
    // then immediately shut down.
    auto release = !event.period();
    if(!is_authorized(event, release)) {
        debug(4, "request registration authorization");
        challenge_message(event);
    }
    else if(!registry::is_extension(event.user())) {
        debug(4, "invalid registration extension {}", event.user());
        event.fail_message(SIP_NOT_FOUND);
    }
    else if(release)
        exec::release_user(event);
    else
        exec::refresh_user(event);
}
} // end anon namespace

event_t::event_t(const exosip::context_t &ctx, const std::string& addr) noexcept :
exosip::event_t(ctx, 500), data_(std::make_shared<data_t>()), authorize_(nullptr), period_(0), refresh_(0), port_(ctx.port())
{
    data_->created = data_->expires = util::steady_clock();
    data_->address = addr;
    data_->schema = ctx.schema();

    if(evt_ && evt_->request) {
        auto msg = evt_->request;
        osip_message_get_authorization(msg, 0, &authorize_);
        if(authorize_ && authorize_->realm) {
            data_->realm = unquote_view(authorize_->realm);
            if(authorize_->username)
                data_->user = unquote_view(authorize_->username);
            if(authorize_->nonce)
                data_->nonce = unquote_view(authorize_->nonce);
            if(authorize_->algorithm)
                data_->algorithm = unquote_view(authorize_->algorithm);
            if(authorize_->message_qop)
                data_->qop = unquote_view(authorize_->message_qop);
            if(authorize_->cnonce)
                data_->client = unquote_view(authorize_->cnonce);
            if(authorize_->nonce_count)
                data_->count = unquote_view(authorize_->nonce_count);
            if(authorize_->response)
                data_->digest = unquote_view(authorize_->response);
        }

        if(data_->algorithm.empty()) {
            if(data_->digest.size() == 32)
                data_->algorithm = "MD5";
            if(data_->digest.size() == 64)
                data_->algorithm = "SHA-256";
        }
        if(msg->req_uri) {
            char *uri = nullptr;
            osip_uri_to_str(msg->req_uri, &uri);
            if(uri) {
                data_->uri = uri;
                exosip::release(uri);
            }
        }

        osip_header_t *header = nullptr;
        osip_message_header_get_byname(msg, "session-expires", 0, &header);
        if(header && header->hvalue)
            period_ = util::get_period(header->hvalue);
        else {
            header = nullptr;
            osip_message_header_get_byname(msg, "expires", 0, &header);
            if(header && header->hvalue)
                period_ = util::get_period(header->hvalue);
        }
        expires(period_ * 1000L);
    }
    if(evt_ && evt_->response) {
        auto msg = evt_->response;
        osip_header_t *header = nullptr;
        osip_message_header_get_byname(msg, "session-expires", 0, &header);
        if(header && header->hvalue)
            refresh_ = util::get_period(header->hvalue);
    }
}

event_t::event_t() noexcept :
data_(std::make_shared<data_t>()), authorize_(nullptr), period_(0), refresh_(0), port_(0)
{
    data_->created = data_->expires = util::steady_clock();
}

auto event_t::content() const -> std::string {
    if(evt_ && evt_->request) {
        auto ctype = osip_message_get_content_type(evt_->request);
        if(!ctype || !ctype->type)
            return "unknown";

        if(ctype->subtype)
            return ctype->type + "/"s + ctype->subtype;

        return ctype->type;
    }
    return "none";
}

void event_t::log_access(int code) const
{
    std::string source = "unknown";
    std::string reason;

    switch(code) {
    case SIP_FORBIDDEN:
        reason = "forbidden";
        break;
    case SIP_406_NOT_ACCEPTABLE:
        reason = "unacceptable";
        break;
    case SIP_ADDRESS_INCOMPLETE:
        reason = "malformed";
        break;
    default:
        return;
    }

    // find source for fail-to-ban uses...
    if(evt_ && evt_->request) {
        auto& vias = evt_->request->vias;
        auto pos = 0;
        auto via = static_cast<osip_via *>(nullptr);

        while(osip_list_eol(&vias, pos) == 0)
            via = static_cast<osip_via_t *>(osip_list_get(&vias, pos++));

        if(via && via->host)
            source = via->host;
    }
    log_error("{}: access failure: {} request", source, reason);
}

auto event_t::remains() const -> unsigned
{
    if(data_->expires <= data_->created)
        return 0;

    auto now = util::steady_clock();
    if(data_->expires <= now)
        return 0;

    return util::steady_duration(now, data_->expires).count();
}

auto event_t::is_expired() const -> bool
{
    if(!evt_)
        return true;
    if(data_->expires == data_->created)
        return false;
    return util::steady_clock() > data_->expires;
}

void event_t::expires(unsigned msecs) {
    if(!evt_)
        return;
    if(msecs > 0)
        data_->expires = data_->created + std::chrono::milliseconds(msecs);
}

void server::allow_local(exosip::msg_t msg)
{
    osip_message_set_header(msg, ALLOW, "OPTIONS,REGISTER,MESSAGE,INVITE,BYE,CANCEL,ACK,INFO,PUBLISH,SUBSCRIBE");
    osip_message_set_header(msg, ACCEPT, "application/sdp,application/dtmf-relay,text/plain,application/pidf+xml");
    osip_message_set_header(msg, ACCEPT_ENCODING, "text/plain");
    osip_message_set_header(msg, ACCEPT_LANGUAGE, "en");    // TODO: Phrasebook
    //osip_message_set_header(msg, SUPPORTED, "100rel");
}

void server::allow_remote(exosip::msg_t msg)
{
    osip_message_set_header(msg, ALLOW, "OPTIONS,MESSAGE,INVITE,BYE,CANCEL,ACK,INFO");
    osip_message_set_header(msg, ACCEPT, "application/sdp,application/dtmf-relay,text/plain");
    osip_message_set_header(msg, ACCEPT_ENCODING, "text/plain");
    osip_message_set_header(msg, ACCEPT_LANGUAGE, "en");    // TODO: Phrasebook
    //osip_message_set_header(msg, SUPPORTED, "100rel");
}

auto server::msg_to_string(exosip::msg_t msg) -> std::string
{
    char *text = nullptr;
    size_t tlen = 0;

    osip_message_to_str(msg, &text, &tlen);
    if(!text)
        return {};

    std::string result(text, tlen);
    exosip::release(text);
    return result;
}

void server::exec_local(exosip::context_t& context, const keyfile& ini)
{
    using namespace exosip;

    keyfile::keys config;
    const auto& sc = ini["server"];
    config.insert(sc.begin(), sc.end());

    const auto& address = config["address"];
    const auto& hostname = config["hostname"];
    const auto names = tokenize(config["names"], " ,;");
    const auto port = stoi(config["port"]);
    auto priority = stoi(config["priority"]);

    realm = config["realm"];
    algorithm = config["algorithm"];
    max_expires = util::get_expires(config["expires"]);

    auto local_address = net::to_address(hostname);
    for(const auto& name : names) {
        localhosts.emplace(name);
        auto addr = net::to_address(name);
        if(addr.empty())
            continue;

        if(local_address.empty())
            local_address = addr;

        localhosts.emplace(addr);
    }

    if(local_address.empty())
        local_address = hostname;

    for(const auto& host : localhosts)
        debug(4, "host: {}\n", host);

    auto prior = util::steady_clock();
    auto expires = prior;

    debug(2, "binding local to {}:{}", address, port);
    context.set_family(ipv6 ? family_type::ipv6 : family_type::ipv4);
    if(!context.listen(address, exosip::listen_type::udp, port))
        log_fail(98, "cannot bind local to {}:{}", address, port);

    if(!proc::set_priority(priority))
        priority = 0;

    log_info("starting {}:{} as local thread: pri={}", address, port, priority);
    for(;;) {
        server::event_t event(context, local_address);
        if(exiting)
            break;

        // handle shutdown operations...
        std::this_thread::yield();
        if(!running) {
            if(!is(event))
                continue;
            switch(event.type()) {
            case EXOSIP_MESSAGE_NEW:
                event.fail_message(SIP_TEMPORARILY_UNAVAILABLE);
                break;
            case EXOSIP_CALL_INVITE:
                event.fail_call(SIP_TEMPORARILY_UNAVAILABLE);
                break;
            default:
                event.automatic();
            }
            continue;
        }

        // fast breakdown of local events without local scopes in switch
        if(is(event)) {
            debug(5, "local: {}", event.text());

            auto request = event.request();
            switch(event.type()) {
            case EXOSIP_MESSAGE_NEW:
                if(!request)
                    event.fail_message(SIP_BAD_REQUEST);
                else if(MSG_IS_OPTIONS(request))
                    do_options(event);
                else if(MSG_IS_REGISTER(request))
                    do_register(event);
                else if(MSG_IS_MESSAGE(request))
                    do_message(event);
                else if(MSG_IS_PUBLISH(request))
                    do_publish(event);
                else
                    event.fail_message(SIP_METHOD_NOT_ALLOWED);
                break;
            case EXOSIP_CALL_INVITE:
                if(!request || event.cid() < 1)
                    event.fail_call(SIP_BAD_REQUEST);
                do_invite(event);
                break;
            case EXOSIP_CALL_MESSAGE_NEW:
                if(MSG_IS_BYE(request))
                    do_bye(event);
                else if(MSG_IS_INFO(request))
                    exec::call_info(event);
                else
                    event.fail_call(SIP_METHOD_NOT_ALLOWED);
                break;
            case EXOSIP_CALL_RINGING:
                exec::ringing_invite(event);
                break;
            case EXOSIP_CALL_CANCELLED:
                exec::release_invite(event);
                break;
            case EXOSIP_CALL_NOANSWER:
                break;
            case EXOSIP_CALL_GLOBALFAILURE:
            case EXOSIP_CALL_SERVERFAILURE:
            case EXOSIP_CALL_REQUESTFAILURE:
                exec::failed_invite(event);
                break;
            case EXOSIP_CALL_ACK:
                exec::ack_invite(event);
                break;
            case EXOSIP_CALL_REINVITE:
                exec::redo_invite(event);
                break;
            case EXOSIP_CALL_ANSWERED:
                exec::answer_invite(event);
                break;
            case EXOSIP_CALL_PROCEEDING:
                exec::trying_invite(event);
                break;
            case EXOSIP_CALL_CLOSED:
                exec::close_invite(event);
                break;
            case EXOSIP_CALL_RELEASED:
                exec::release_invite(event);
                break;
            default:
                event.automatic();
            }
        }

        // make sure even on high load we call automatic ops ~twice a second
        auto now = util::steady_clock();
        if(!event || now > (prior + 500ms)) {
            debug(6, "local events automatic");
            prior = event.created();
            context.automatic();
        }

        // time tracking for expiring issued nonces...
        if(now >= (expires + 15000ms) && !nonces.empty()) {
            debug(5, "nonce expiration");
            expires = now;
            auto expired = now - 5000ms;
            for(auto it = nonces.begin(); it != nonces.end();) {
                if(it->created() < expired) {
                    debug(6, "nonce expires {}", it->hex());
                    it = nonces.erase(it);
                }
                else
                    ++it;
            }
        }
    }

    log_info("stopping local events thread");
}

void server::exec_remote(exosip::context_t& context, const keyfile& ini)
{
    using namespace exosip;

    keyfile::keys config;
    const auto& sc = ini["server"];
    const auto& rc = ini["remote"];
    config.insert(rc.begin(), rc.end());
    config.insert(sc.begin(), sc.end());


    const auto& upstream = config["upstream"];
    const auto& address = config["address"];
    const auto port = stoi(config["port"]);

    if(upstream.empty()) {
        log_info("remote adapter disabled");
        return;
    }

    auto remote_address = net::to_address(address);
    if(remote_address.empty())
        remote_address = address;

    debug(2, "binding remote to {}:{}", address, port);
    context.set_family(ipv6 ? family_type::ipv6 : family_type::ipv4);
    context.srv_lookup(true);
    if(!context.listen(address, listen_type::tcp, port))
        log_fail(98, "cannot bind remote to {}:{}", address, port);

    auto prior = util::steady_clock();
    log_info("starting {}:{} as remote thread", address, port);
    for(;;) {
        server::event_t event(context, remote_address);
        if(!running)
            break;

        // fast breakdown of remote events without local scopes in switch
        if(is(event)) {
            debug(5, "remote: {}", event.text());

            auto request = event.request();
            switch(event.type()) {
            case EXOSIP_MESSAGE_NEW:
                if(!request)
                    event.fail_message(SIP_BAD_REQUEST);
                else if(MSG_IS_OPTIONS(request))
                    server_options(event, [](exosip::msg_t msg){
                        allow_remote(msg);
                    });
                else
                    event.fail_message(SIP_METHOD_NOT_ALLOWED);
                break;
            default:
                event.automatic();
            }
        }

        // make sure even on high load we call automatic ops ~twice a second
        if(!event || util::steady_clock() > (prior + 500ms)) {
            debug(6, "remote events automatic");
            prior = event.created();
            context.automatic();
        }
    }

    log_info("stopping remote events thread");
}
