/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef EVENTS_HPP
#define EVENTS_HPP

#include "main.hpp"
#include "clocking.hpp"

#include <exosip.hpp>

class keyfile;

namespace server {
class event_t : public exosip::event_t
{
public:
    explicit event_t(const exosip::context_t &ctx, const std::string& addr) noexcept;
    event_t() noexcept;

    inline explicit operator eXosip_t *() const {
        return ctx_;
    }

    inline auto expires() const {
        if(data_->expires == data_->created)
            return steady_time::max();
        return data_->expires;
    }

    inline void fail_message(int code, osip_message_t *msg = nullptr) const {
        log_access(code);
        exosip::guard_t lock(context());
        eXosip_message_send_answer(context(), tid(), code, msg);
    }

    inline void fail_call(int code, osip_message_t *msg = nullptr) const {
        log_access(code);
        exosip::guard_t lock(context());
        eXosip_call_send_answer(context(), tid(), code, msg);
    }

    inline auto created() const {return data_->created;}
    inline auto address() const {return data_->address;}
    inline auto schema() const {return data_->schema;}
    inline auto realm() const {return data_->realm;}
    inline auto user() const {return data_->user;}
    inline auto uri() const {return data_->uri;}
    inline auto authorize() const {return authorize_;}
    inline auto algorithm() const {return data_->algorithm;}
    inline auto digest() const {return data_->digest;}
    inline auto nonce() const {return data_->nonce;}
    inline auto count() const {return data_->count;}
    inline auto client() const {return data_->client;}
    inline auto qop() const {return data_->qop;}
    inline auto period() const {return period_;}
    inline auto refresh() const {return refresh_;}
    inline auto port() const { return port_;}
    inline auto from() const { return (evt_ && evt_->request) ? evt_->request->from : nullptr;}
    inline auto to() const { return (evt_ && evt_->request) ? evt_->request->to : nullptr;}

    inline auto subject() const {
        osip_header_t *header{nullptr};
        if(evt_ && evt_->request)
            osip_message_get_subject(evt_->request, 0, &header);
        return (header && header->hvalue) ? std::string_view(header->hvalue) : "";
    }

    inline auto publish() const {
        osip_header_t *header{nullptr};
        if(evt_ && evt_-> request)
            osip_message_header_get_byname(evt_->request, "event", 0, &header);
        return (header && header->hvalue) ? std::string_view(header->hvalue) : "";
    }

    inline auto agent() const {
        osip_header_t *header{nullptr};
        if(evt_ && evt_->request)
            osip_message_get_user_agent(evt_->request, 0, &header);
        return (header && header->hvalue) ? std::string_view(header->hvalue) : "";
    }

    inline auto body() const {
        osip_body_t *data{nullptr};
        if(evt_ && evt_->request)
            osip_message_get_body(evt_->request, 0, &data);
        return data;
    }

    inline auto reply() const {
        osip_body_t *data{nullptr};
        if(evt_ && evt_->response)
            osip_message_get_body(evt_->response, 0, &data);
        return data;
    }

    auto content() const -> std::string;
    auto is_expired() const -> bool;
    auto remains() const -> unsigned;
    void log_access(int code) const;
    void expires(unsigned msecs);

private:
    struct data_t final {
        steady_time created, expires;
        std::string address, schema, uri;
        std::string_view realm, user, nonce, algorithm, digest, qop, client, count;
    };

    std::shared_ptr<data_t> data_;
    osip_authorization_t *authorize_;
    time_t period_, refresh_;
    uint16_t port_;
};

inline auto content_view(osip_body_t *body) {
    if(body && body->body)
        return std::string_view(body->body, body->length);
    return std::string_view();
}

void allow_local(exosip::msg_t msg);
void allow_remote(exosip::msg_t msg);
auto content_type(osip_body_t *body) -> std::string;
auto msg_to_string(exosip::msg_t msg) -> std::string;

void exec_local(exosip::context_t& context, const keyfile& ini);
void exec_remote(exosip::context_t& context, const keyfile& ini);
} // end namespace

/*!
 * Contexts and server isolated exosip event threads.
 * \file events.hpp
 */
#endif
