/*
 * Copyright (C) 2017-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef IPC_HPP
#define IPC_HPP

#include <cstdint>
#include <ctime>
#include <atomic>

#include <unistd.h>
#include <sys/socket.h>

inline constexpr auto IPC_ID_MAX   = 32;
inline constexpr auto IPC_SMS_MAX  = 160;
inline constexpr auto IPC_SUBJ_MAX = 64;
inline constexpr auto IPC_VERSION  = 0;
inline constexpr auto IPC_KEY_SIZE = 32;
inline constexpr auto IPC_SET_SIZE = 64;

/* Known IPC requests */
typedef enum {  // NOLINT
    IPC_STATUS = 1,
    IPC_MESSAGE,
    IPC_LEVEL,
    IPC_RELOAD,
    IPC_ACTIVE,
    IPC_ASSIGN,
    IPC_CLEARS,
    IPC_PRESENCE,
} ipc_type_t;

/* IPC mqueue messages */
struct ipc_msg {
    ipc_type_t type;
    pid_t pid;
    uint8_t ver;
    uint8_t res0;
    union {
        uint8_t level;
        struct {
            char to[IPC_ID_MAX + 1];
            char text[IPC_SMS_MAX + 1];
            char subj[IPC_SUBJ_MAX + 1];
        } message;
        struct {
            unsigned ext;
            char key[IPC_KEY_SIZE];
            char value[IPC_SET_SIZE];
        } registry;
    } body;
};

/* Server registry for each extension */
struct ipc_reg {
    std::atomic<unsigned> count;
    time_t activated, expires;       /* 0 if inactive */
    struct sockaddr_storage address;
    uint16_t lines;
    volatile enum {GONE = 0, HERE, AWAY, DND} presence;
};

/* Server system info */
struct ipc_sys {
    time_t started;
    pid_t pid;                      /* server pid */
    uint8_t version;
    uint8_t level;
    uint16_t first;
    char state[16];
    char realm[64];
    ipc_reg trunk;
};

/* Each server call; mapped to call capacity */
struct ipc_call {
    uint64_t id;                    /* 0 if inactive */
    time_t created;
    uint16_t active, ringing, segment;
    char caller[64], dialed[64], remote[128];
    enum {INVITING, RINGING, CONNECTED, CLOSING} state;
    enum {LOCAL, INBOUND, OUTGOING, NONE} type;
};

/*!
 * Use coventry ipc services. IPC includes shared memory blocks for the
 * registry and call processing, and a control mqueue. This header may be
 * be possible to include in stand-alone external third party projects
 * as well. Because ipc does not cross multiple machines, native byte order
 * and struct alignment can be used.
 * \file ipc.hpp
 */
#endif
