/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "messages.hpp"
#include "clocking.hpp"
#include "logger.hpp"
#include "events.hpp"
#include "utils.hpp"

#include <keyfile.hpp>

#include <functional>
#include <utility>
#include <thread>
#include <condition_variable>
#include <fstream>
#include <iomanip>

using namespace server;

namespace {
struct args_t final {
    args_t() = default;
    args_t(keyfile keys) : config(std::move(keys)) {}
    args_t(message m) : msg(std::move(m)) {}
    args_t(const logmsg& l) : log(std::make_shared<logmsg>(l)) {}

    std::shared_ptr<logmsg> log;
    message msg;
    keyfile config;
};

using actor_t = void (*)(const args_t&);
using msg_t = std::pair<actor_t, args_t>;

time_t max_expires = 30;
bool stopping = false;
std::ofstream logfile, cdrfile;
std::list<msg_t> msgs;
std::mutex msglock;
std::condition_variable cvar;

void do_shutdown([[maybe_unused]] const args_t& args)
{
    stopping = true;
}

void do_reload(const args_t& args)
{
    const auto& keys = args.config;
    auto config = keys.at("messages");
    max_expires = util::get_expires(config["expires"]);

    // Allows for log rotation...
    if(logfile.is_open())
        logfile.close();
    logfile.open(SERVER_LOGGER_PATH "/coventry.log", std::ios_base::app); // FlawFinder: ignore

    if(cdrfile.is_open())
        cdrfile.close();
    cdrfile.open(SERVER_LOGGER_PATH "/calls.log", std::ios_base::app); // FlawFinder: ignore
}

void do_local(const args_t& args)
{
    const auto& send = args.msg;
    auto context = send.context();
    auto to = send.to();
    auto type = send.type();
    auto body = send.body();
    auto from = send.from();
    auto subject = send.subject();

    debug(4, "deliver to {}", send.route());

    std::this_thread::yield();

    osip_message_t *msg = nullptr;
    exosip::guard_t lock(context);
    auto result = eXosip_message_build_request(context, &msg, "MESSAGE", to.c_str(), from.c_str(), send.route());
    if(msg) {
        if (!subject.empty())
            osip_message_set_subject(msg, subject.c_str());
        osip_message_set_body(msg, body.data(), body.size());
        osip_message_set_content_type(msg, type.c_str());
        allow_local(msg);
        eXosip_message_send_request(context, msg);
    }
    else
        log_error("failed to send message to {}; error={}", send.to(), result);
}

void do_logfile(const args_t& args)
{
    const auto& log = *args.log;
    if(!logfile.is_open())
        return;

    struct tm now = util::local_time(log.created);
    logfile << std::put_time(&now, "%Y-%m-%d %X ") << log.msg << std::endl;
}

void do_cdrfile(const args_t& args)
{
    const auto& cdr = *args.log;
    if(!cdrfile.is_open())
        return;

    auto duration = print_format("{}d{}h{}m{}s", cdr.duration / 86400, (cdr.duration / 3600) % 24, (cdr.duration / 60) % 60, cdr.duration % 60);
    std::stringstream json;
    json <<  std::put_time(gmtime(&cdr.created), "{\"m\":\"cdr\",\"v\":0,\"p\":\"sip\",\"s\":\"coventry/" PROJECT_VERSION "\",\"zt\":\"%FT%TZ\",") << cdr.msg << R"(,"td":")" << duration  << "\"}";
    cdrfile << json.str() << std::endl;
}

void dispatch(actor_t actor, const args_t& args = args_t())
{
    std::unique_lock<std::mutex> lock(msglock);
    if(stopping)
        return;

    msgs.emplace_back(actor, args);
    lock.unlock();
    cvar.notify_one();
}
} // end anon namespace

message::message(std::string_view from_name, std::string_view to_uri, std::string_view route_to, std::string_view msg_subject, std::string_view msg_text, std::string_view msg_type, eXosip_t *context) : data_(std::make_shared<data>(from_name, to_uri, route_to, msg_subject, msg_text, msg_type)), ctx_(context)
{
    time(&data_->expires);
    data_->expires += max_expires;
}

void server::logger(const std::string& msg)
{
    exec::post_log(logmsg(msg));
}

void server::exec_messages(const keyfile& config)
{
    auto args = args_t(config);

#ifdef DEBUG
    std::remove(SERVER_LOGGER_PATH "/coventry.log");
    std::remove(SERVER_LOGGER_PATH "/calls.log");
#endif

    do_reload(args);
    log_info("starting messages thread");
    while(!stopping) {
        std::unique_lock<std::mutex> lock(msglock);
        if(!msgs.empty()) {
            auto copy = std::move(msgs.front());
            auto [actor, args] = copy;
            msgs.pop_front();
            lock.unlock();
            try {
                actor(args);
            }
            catch(const std::exception& e) {
                log_error("registry actor error: {}", e.what());
            }
            continue;
        }
        cvar.wait(lock);
    }
    log_info("stopping messages thread");
}

void exec::stop_messages()
{
    dispatch(&do_shutdown);
}

void exec::reload_messages(const keyfile& config)
{
    dispatch(&do_reload, config);
}

void exec::deliver_local(const message& msg)
{
    dispatch(&do_local, msg);
}

void exec::post_log(const logmsg& log)
{
    dispatch(&do_logfile, log);
}

void exec::post_cdr(const logmsg& cdr)
{
    dispatch(&do_cdrfile, cdr);
}
