/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.hpp"
#include "proc.hpp"
#include "utils.hpp"
#include "logger.hpp"

#include <strings.hpp>

#include <csignal>
#include <cstring>
#include <mutex>

#include <pthread.h>
#include <sched.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#ifndef _PATH_TTY
#define _PATH_TTY   "/dev/tty"  // NOLINT
#endif

using namespace server;
using namespace std::literals::string_literals;

namespace {
inline constexpr int max_fd = sizeof(fd_set) * 8;
std::mutex forklock;
}

void proc::ipc_notify(pid_t pid, bool success)
{
    if(!pid || pid == getpid())
        return;

    // control who is signalled when root user
    char path[48];
    char target[128];
    snprintf(path, sizeof(path), "/proc/%lld/exe", static_cast<long long>(pid));
    auto result = readlink(path, target, sizeof(target) - 1); // FlawFinder: protected proc names
    if (result < 5) {
        log_error("notify unknown or invalid pid {}", pid);
        return;
    }
    target[result] = 0;
    auto cp = strrchr(target, '/');
    if(!cp || strncmp(cp, "/pbx-", 5) != 0) {
        log_error("target pid not a pbx utility {}", target);
        return;
    }

    if(success)
        kill(pid, SIGUSR1);
    else
        kill(pid, SIGUSR2);
}

auto proc::detach(std::string argv0, std::list<std::string> args, std::list<std::string> env) -> pid_t
{
    std::lock_guard<std::mutex> lock(forklock);
    auto pid = fork();
    if(pid != 0)
        return pid;

    ::signal(SIGQUIT, SIG_DFL);
    ::signal(SIGTERM, SIG_DFL);
    ::signal(SIGINT, SIG_DFL);
    ::signal(SIGCHLD, SIG_DFL);
    ::signal(SIGPIPE, SIG_DFL);
    ::signal(SIGHUP, SIG_DFL);
    ::signal(SIGALRM, SIG_DFL);
    ::signal(SIGABRT, SIG_DFL);
    ::signal(SIGUSR1, SIG_DFL);

#ifdef  SIGTTOU
    ::signal(SIGTTOU, SIG_IGN); // NOLINT
#endif

#ifdef  SIGTTIN
    ::signal(SIGTTIN, SIG_IGN); // NOLINT
#endif

#ifdef  SIGTSTP
    ::signal(SIGTSTP, SIG_IGN); // NOLINT
#endif

    for(int fd = 3; fd < max_fd; ++fd)
        ::close(fd);

    if(setpgid(0, getpid()) == -1)
        ::quick_exit(EXIT_FAILURE);

#if defined(SIGTSTP) && defined(TIOCNOTTY)
    auto fd = ::open(_PATH_TTY, O_RDWR);    // FlawFinder: System file
    if(fd >= 0) {
        ioctl(fd, TIOCNOTTY, NULL);
        close(fd);
    }
#else
    if(getppid() != 1) {
        if((pid = fork()) < 0)
            ::quick_exit(EXIT_FAILURE);
        else if(pid > 0)
            ::quick_exit(EXIT_SUCCESS);
    }
#endif

    // construct child environment
    env.emplace_back(print_format("PATH={}:/usr/local/bin:/usr/pkg/bin:/usr/bin:/bin", SERVER_LIBEXEC_PATH));
    env.emplace_back(print_format("LANG={}", util::get_env("LANG")));
    env.emplace_back("HOME=" SERVER_PREFIX_PATH);
    env.emplace_back("SERVER_REALM="s + util::get_env("REALM", 64));
    env.emplace_back("SERVER_STATE="s + util::get_env("STATE", 64));
    env.emplace_back("SERVER_CONFIG=" SERVER_CONFIG_FILE);
    env.emplace_back("SERVER_LIBEXEC=" SERVER_LIBEXEC_PATH);
    env.emplace_back("SERVER_TYPE=COVENTRY");
    env.emplace_back("SERVER_API=1");
    ::chdir(SERVER_PREFIX_PATH);

    char *argv[args.size() + 2];
    char *envp[env.size() + 1];

    unsigned argc = 1;
    while(!args.empty()) {
        argv[argc++] = ::strdup(args.front().c_str());
        args.pop_front();
    }

    unsigned envc = 0;
    while(!env.empty()) {
        envp[envc++] = ::strdup(env.front().c_str());
        env.pop_front();
    }

    if(!getuid())
        setuid(nobody);

    argv0 = SERVER_LIBEXEC_PATH "/" + argv0;
    argv[0] = ::strdup(argv0.c_str());
    argv[argc] = nullptr;
    envp[envc] = nullptr;
    execve(argv[0], argv, envp);
    ::quick_exit(EXIT_FAILURE);
}

auto proc::get_priority() -> int
{
    auto tid = pthread_self();
    struct sched_param sp{};
    int policy{};

    if(pthread_getschedparam(tid, &policy, &sp))
        return 0;

#ifdef  SCHED_BATCH
    if(policy == SCHED_BATCH)
        return -1;
#endif

    return sp.sched_priority;
}

auto proc::set_priority(int priority) -> bool
{
    auto tid = pthread_self();
    struct sched_param sp{};
    int policy{};

    std::memset(&sp, 0, sizeof(sp));

    policy = SCHED_OTHER;
#ifdef  SCHED_FIFO
    if(priority > 0) {
        policy = SCHED_FIFO;
        priority = sched_get_priority_min(policy) + priority - 1;
        if(priority > sched_get_priority_max(policy))
            priority = sched_get_priority_max(policy);
    }
#endif
#ifdef  SCHED_BATCH
    if(priority < 0) {
        policy = SCHED_BATCH;
        priority = 0;
    }
#endif
    if(policy == SCHED_OTHER)
        priority = 0;

    sp.sched_priority = priority;
    return pthread_setschedparam(tid, policy, &sp) == 0;
}
