/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "events.hpp"
#include "registry.hpp"
#include "calls.hpp"
#include "logger.hpp"
#include "messages.hpp"
#include "network.hpp"
#include "utils.hpp"
#include "proc.hpp"

#include <strings.hpp>
#include <keyfile.hpp>

#include <condition_variable>
#include <utility>
#include <functional>

#include <fcntl.h>
#include <sys/mman.h>

using namespace std::literals::string_literals;
using namespace server;

namespace {
inline constexpr auto reg_entries = CONFIG_REG_ENTRIES;
inline constexpr auto reg_size = sizeof(ipc_reg) * reg_entries;
inline constexpr auto mem_size = reg_size + sizeof(ipc_sys);
inline constexpr auto help_text = "*** system commands: /echo /msg /who";

struct args_t final {
    args_t() = default;
    args_t(keyfile cfg) : config(std::move(cfg)) {}
    args_t(event_t evt) : event(std::move(evt)) {}
    args_t(ipc_var evt) : ipc(std::move(evt)) {}

    keyfile config;
    event_t event;
    ipc_var ipc;
};

using actor_t = void (*)(const args_t&);
using msg_t = std::pair<actor_t, args_t>;

std::unordered_map<std::string, registry> clients;
std::array<registry, reg_entries> mapped;
registry empty_registry;
std::string hostname = "localhost";
std::string realm = "Coventry";
std::string state = "none";
std::string sitename = "Coventry Server";
std::string sitefrom = "sip:system@locahost:5060";
uint16_t hostport = 5060;
volatile bool timer = false;
ipc_reg *reg_shared = nullptr;
ipc_sys *reg_system = nullptr;
keyfile::keys server_messages;
time_t max_expires = 300;
keyfile key_storage;
auto key_updates = true;    // always save after initial load
auto dialplan_min = CONFIG_DIALPLAN;
auto dialplan_max = dialplan_min + reg_entries - 1U;
auto dialplan_ext = ext_t(CONFIG_DIALPLAN);
auto stopping = false;

std::list<msg_t> msgs;
std::mutex msglock;
std::condition_variable cvar;

void dispatch(actor_t actor, const args_t& args = args_t())
{
    std::unique_lock<std::mutex> lock(msglock);
    if(stopping)
        return;

    msgs.emplace_back(actor, args);
    lock.unlock();
    cvar.notify_one();
}

class registry_map final
{
public:
    inline registry_map() { // NOLINT
        memset(data_, 0, sizeof(data_));
    }

    explicit inline registry_map(std::string_view list) { // NOLINT
        memset(data_, 0, sizeof(data_));
        add(list);
    }

    auto operator[](ext_t index) const -> bool {
        if(index < ext_t(dialplan_min) || index > ext_t(dialplan_max))
            throw std::range_error("outside extension map");

        index -= dialplan_min;
        uint8_t mask = 0x01 << index % 8;
        return data_[index >> 3] & mask;
    }

    auto operator[](const registry_map& map) const {
        for(auto index = 0UL; index < sizeof(data_); ++index) {
            if((data_[index] & map.data_[index]) != map.data_[index])
                return false;
        }
        return true;
    }

    inline auto operator+=(const ext_t value) -> registry_map& {
        operator()(value, true);
        return *this;
    }

    inline auto operator-=(const ext_t value) -> registry_map& {
        operator()(value, false);
        return *this;
    }

    inline auto operator==(const ext_t value) const {
        return operator[](value);
    }

    inline auto operator!=(const ext_t value) const {
        return !operator[](value);
    }

    inline auto operator()(ext_t ext) const {
        return operator[](ext);
    }

    inline auto data() const {
        return &data_;
    }

    auto operator&=(const registry_map& from) {
        for(auto index = 0UL; index < sizeof(data_); ++index)
            data_[index] &= from.data_[index];
        return *this;
    }

    auto operator|=(const registry_map& from) {
        for(auto index = 0UL; index < sizeof(data_); ++index)
            data_[index] |= from.data_[index];
        return *this;
    }

    auto operator^=(const registry_map& from) {
        for(auto index = 0UL; index < sizeof(data_); ++index)
            data_[index] ^= from.data_[index];

        return *this;
    }

    void operator()(ext_t index, bool state) {
        if(index < ext_t(dialplan_min) || index > ext_t(dialplan_max))
            throw std::range_error("outside extension map");

        index -= dialplan_min;
        uint8_t mask = 0x01 << index % 8;
        if(state)
            data_[index >> 3] |= mask;
        else
            data_[index >> 3] &= ~mask;
    }

    auto is_sync(const ext_t& index) const {
        if(index < ext_t(dialplan_min) || index > ext_t(dialplan_max))
            return false;

        auto offset = *index - dialplan_min;
        uint8_t mask = 0x01 << (offset % 8);

        offset >>= 3;
        auto byte = data_[offset];
        while(byte != data_[offset])
            byte = data_[offset];

        return (byte & mask) != 0;
    }

    auto count() const {
        auto count = 0U;
        auto index = dialplan_min;
        while(index <= dialplan_max) {
            auto offset = index - dialplan_min;
            auto mask = 0x01U;
            auto byte = data_[offset / 8];
            if(!byte) { // skip empty optimization...
                index += 8;
                continue;
            }
            for(auto bits = 0; bits < 8; ++bits) {
                if(byte & mask)
                    ++count;

                ++index;
                mask <<= 1;
            }
        }
        return count;
    }

    auto apply(std::function<bool(registry&)> pred) const {
        auto total = 0U;
        auto index = dialplan_min;
        while(index <= dialplan_max) {
            auto offset = index - dialplan_min;
            auto mask = 0x01U;
            auto byte = data_[offset / 8];
            if(!byte) { // skip empty optimization...
                index += 8;
                continue;
            }
            for(auto bits = 0; bits < 8; ++bits) {
                if(byte & mask) {
                    auto& entry = mapped.at(offset + bits);
                    if(is(entry) && pred(entry))
                        ++total;
                }
                ++index;
                mask <<= 1;
            }
        }
        return total;
    }

    inline auto empty() const {
        return std::all_of(data_, data_ + sizeof(data_), [](auto bits) {
            return !bits;
        });
    }

    inline operator bool() const {
        return !empty();
    }

    inline auto operator!() const {
        return empty();
    }

    inline auto clear() {
        memset(data_, 0, sizeof(data_));
    }

    void add(std::string_view list) {
        auto tokens = tokenize(list, ",:; ");
        for(const auto& token : tokens) {
            auto value = util::get_value(token) - dialplan_min;
            if(value < 0 || value >= reg_entries)
                continue;
            auto mask = 0x01 << (value % 8);
            value /= 8;
            data_[value] |= mask;
        }
    }

    inline static auto size() {
        return sizeof(data_);
    }

private:
    uint8_t data_[reg_entries / 8];
};

auto operator&(const registry_map& l, const registry_map& r) {
    registry_map tmp(l);
    tmp &= r;
    return tmp;
}

[[maybe_unused]] auto operator|(const registry_map& l, const registry_map& r) {
    registry_map tmp(l);
    tmp |= r;
    return tmp;
}

auto operator^(const registry_map& l, const registry_map& r) {
    registry_map tmp(l);
    tmp ^= r;
    return tmp;
}

// registry maps we use...

registry_map active_map;
registry_map coverage_map[reg_entries];
registry_map delayed_map[reg_entries];
std::unordered_map<std::string, registry_map> group_map;

// helper functions...

void sync_storage()
{
    if(!key_updates)
        return;
    key_storage.write("registry.keys");
    key_updates = false;
}

auto system_from()
{
    return print_format("\"{}\" <{}>", sitename, sitefrom);
}

auto is_authorized(registry& entry, const event_t& event)
{
    auto keys = entry.keys();
    if(!entry.verify(event))
        return false;

    try {
        if(event.algorithm() == "SHA-256")
            return util::sha256_digest(keys.at("sha256"), event) == event.digest();
        if(event.algorithm() == "MD5")
            return util::md5_digest(keys.at("md5"), event) == event.digest();
    }
    catch(...) {}

    log_warn("User {} has no {} secret", event.user(), event.algorithm());
    return false;
}

auto chk_secret(std::string_view user, std::string_view secret, size_t size)
{
    if(size != secret.size())
        return false;

    if(std::any_of(secret.cbegin(), secret.cend(), [](char ch){
        return ch < '0' || ch > 'f' || (ch > '9' && ch < 'a');
    })) {
        log_warn("user {} invalid secret {}", user, secret);
        return false;
    }
    return true;
}

void notify()
{
    timer = true;
    cvar.notify_one();
}

auto find(std::string_view user) noexcept -> registry&
{
    if(user.empty())
        return empty_registry;

    auto value = 0U;
    for(auto ch : user) {
        if(!isdigit(ch))
            return empty_registry;
        value = (value * 10) + (ch - '0');
    }

    if(value < dialplan_min || value > dialplan_max)
        return empty_registry;

    return mapped.at(value - dialplan_min);
}

[[maybe_unused]] auto find(const event_t& event) noexcept -> registry& //TODO: use this?
{
    if(event.authorize())
        return find(event.user());

    contact_t route(event);
    try {
        return clients.at(route.addr());
    }
    catch(...) {
        return empty_registry;
    }
}

void changed(const registry_map& current)
{
    auto prior = active_map;
    active_map = current;

    auto changes = current ^ prior;
    auto dropped = changes & prior;
    //auto added = changes & current;

    for(auto base = 0UL; base < mapped.size(); ++base) {
        auto ext = dialplan_ext + base;
        if(!dropped[ext])
            continue;
        debug(4, "registry: drop {}", ext);
        mapped.at(base).drop();
    }
}

auto rescan()   // fast sync compute...
{
    registry_map current;
    time_t now{}, latest = 0;
    time(&now);

    debug(2, "registry timeout scan");
    sync_storage();
    for(auto& entry : mapped) {
        auto expires = entry.expires();
        if(!expires)            // if already expired, ignore
            continue;
        if(expires <= now) {     // drop expired entry...
            debug(4, "registry: expire {}", entry.extension());
            auto& shared = reg_shared[entry.extension() - dialplan_min];
            shared.activated = shared.expires = 0;
            entry.drop();
            continue;
        }
        current += entry.extension();
        if(expires < latest)
            latest = expires;
    }

    // locked registry
    changed(current);
    return latest;
}

// cli message parser...

auto cli_command(const registry& user, std::string_view text)
{
    auto tokens = tokenize(text);
    if(tokens[0] == "/echo") {
        text.remove_prefix(5);
        while(!text.empty() && text.front() == ' ')
            text.remove_prefix(1);
        exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", text, "text/plain", user.context()));
        return true;
    }

    if(tokens[0] == "/msg") {
        if(tokens.size() < 3) {
            exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", "*** either no service named or no text given", "text/plain", user.context()));
            return true;
        }
        auto argv0 = "msg_" + std::string{tokens[1]};
        auto to = std::to_string(user.extension());
        auto filename = SERVER_LIBEXEC_PATH "/" + argv0;

        if(access(filename.c_str(), R_OK)) { // FlawFinder: these are locked
            exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", "*** invalid service named", "text/plain", user.context()));
            return true;
        }

        std::list<std::string> args, env;
        args.push_back(to);
        for(auto index = 2U; index < tokens.size(); ++index)
            args.emplace_back(std::string{tokens[index]});

        env.emplace_back(print_format("CALLER_ID={}", user.display()));
        proc::detach(argv0, args, env);
        return true;
    }

    if(tokens[0] == "/who") {
        auto id = std::to_string(user.extension());
        std::string result, status;
        if(tokens.size() == 1)
            tokens[0] = id;
        else
            tokens.erase(tokens.begin());
        for(auto token : tokens) {
            if(registry::is_extension(token)) {
                auto& entry = find(token);
                if (is(entry)) {
                    status = "here";
                    switch(entry.presence()) {
                    case ipc_reg::GONE:
                        status = "gone";
                        break;
                    case ipc_reg::DND:
                        status = "dnd";
                        break;
                    case ipc_reg::AWAY:
                        status = "away";
                        break;
                    case ipc_reg::HERE:
                        if(entry.shm()->count)
                            status = "busy";
                        break;
                    }
                }
                else
                    status = "gone";
                result = print_format("Extension {} is {}, status={}, agent={}", entry.extension(), entry.contact().display(), status, entry.agent());
            }
            else
                result = print_format("*** Unknown {}", token);
            exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", result, "text/plain", user.context()));
        }
        return true;
    }
    return false;
}

auto cli_help(const registry& user)
{
    exec::deliver_local(message(system_from(), user.origin_uri(), user.route_to(), "", help_text, "text/plain", user.context()));
    return true;
}

// our actors...

void do_shutdown([[maybe_unused]] const args_t& args)
{
    stopping = true;
    exec::stop_messages();
    exec::stop_calls();
}

void do_reload(const args_t& args)
{
    const auto& keys = args.config;
    auto common = keys.at("common");
    auto server = keys.at("server");
    auto groups = keys.at("groups");
    server_messages = keys.at("messages");
    realm = server["realm"];
    hostname = server["hostname"];
    hostport = stoi(server["port"]);
    sitefrom = "sip:system@" + hostname + ":" + std::to_string(hostport);
    sitename = server["sitename"];
    state = server["state"];
    max_expires = util::get_expires(server["expires"]);
    uint16_t base = 0;
    time_t now{};

    setenv("STATE", state.c_str(), 1);
    setenv("REALM", realm.c_str(), 1);

    sync_storage();
    registry_map current;
    time(&now);
    debug(4, "registry: reload for realm {} state {}", realm, state);
    util::set_string(reg_system->state, sizeof(reg_system->state), state.c_str());
    util::set_string(reg_system->realm, sizeof(reg_system->realm), realm.c_str());

    // clear coverage maps...
    for(auto pos = 0U; pos < reg_entries; ++pos) {
        coverage_map[pos].clear();
        delayed_map[pos].clear();
    }

    // clear and load our named groups...
    group_map.clear();
    for(const auto& [key, value] : groups)
        group_map.emplace(key, value);
    group_map.try_emplace("system", "");    // add system group if missing...

    // load and activate extension entries
    for(auto& entry : mapped) {
        auto pos = base;
        auto ext = dialplan_ext + (base++);
        auto key = std::to_string(ext);
        auto prior = !entry.empty();
        auto exists = keys.exists(key);

        // prune dead keys
        if(!exists && key_storage.exists(key)) {
            key_storage.remove(key);
            key_updates = true;
        }

        if(entry.empty() && !exists)
            continue;

        if(entry.empty())
            entry = registry(ext);

        if(keys.exists(key)) {
            if(prior)
                debug(4, "registry: update {}", ext);
            else
                debug(4, "registry: create {}", ext);
            auto pairs = keys.at(key);
            pairs.insert(common.begin(), common.end());
            try {
                auto stored = key_storage.at(key);
                for(const auto& [id, value] : stored)
                    pairs.insert_or_assign(id, value);
            }
            catch(...) {}
            entry.keys(pairs);

            try {
                auto coverage = pairs.at("coverage");
                coverage_map[pos].add(coverage);
            }
            catch(...) {
                coverage_map[pos] += entry.extension();
            }

            try {
                auto delayed = pairs.at("delayed");
                delayed_map[pos].add(delayed);
            }
            catch(...) {}
        }
        else {
            if(prior)
                debug(4, "registry: remove {}", ext);

            entry = empty_registry;
        }

        if(entry.expires() > now)
            current += ext;
    }
    changed(current);
}

void do_refresh(const args_t& args)
{
    const auto& event = args.event;
    auto& entry = find(event.user());
    auto& shared = reg_shared[entry.extension() - dialplan_min];
    auto period = event.period();
    auto agent = event.agent();

    if(period < 60)
        period = 60;
    else if(period > max_expires)
        period = max_expires;

    if(entry.empty()) {
        event.fail_message(SIP_NOT_FOUND);
        return;
    }

    if(!is_authorized(entry, event)) {
        log_error("failed to authorize {}", event.user());
        event.fail_message(SIP_FORBIDDEN);
        return;
    }

    auto prior_to = entry.remote();
    auto registered = entry.refresh(event, period);
    if(registered) {
        if(event.context() != entry.context()) {
            log_warn("invalid context for user {}", event.user());
            event.fail_message(SIP_FORBIDDEN);
            return;
        }
        debug(4, "refreshing user {} for {} seconds", event.user(), period);
        if(entry.remote() != prior_to) {
            debug(5, "changing user {} from {} to {}", entry.extension(), prior_to, entry.remote());
            clients.erase(prior_to);
            clients[entry.remote()] = entry;
        }
    }
    else {
        auto contact = entry.origin();
        auto pairs = entry.keys();
        auto presence = pairs.at("presence");

        active_map += entry.extension();
        net::store(shared.address, contact.host(), contact.port());
        time(&shared.activated);

        if(presence == "dnd")
            shared.presence = ipc_reg::DND;
        else if(presence == "gone" || presence == "off" || presence == "closed")
            shared.presence = ipc_reg::GONE;
        else if(presence == "away")
            shared.presence = ipc_reg::AWAY;
        else
            shared.presence = ipc_reg::HERE;

        log_notice("activating user {} for {} seconds from {}", event.user(), period, agent);
        notify();
        debug(5, "adding user {} to {}", entry.extension(), entry.remote());
        clients[entry.remote()] = entry;
        if(entry.welcome())
            system_message(server_messages.at("welcome"));
    }
    shared.expires = entry.expires();
    debug(4, "registering from {} with {}", entry.origin_uri(), event.algorithm());
    entry.receipt(event, period);
}

void do_release(const args_t& args)
{
    const auto& event = args.event;
    auto& entry = find(event.user());

    if(entry.empty()) {
        event.fail_message(SIP_NOT_FOUND);
        return;
    }

    nonce_t nonce(event.nonce());
    if(nonce != entry.nonce()) {
        debug(4, "release expected {}, got {}", entry.nonce().hex(), nonce.hex());
        event.fail_message(SIP_FORBIDDEN);  // TODO: challenge message??
        // this means we could not verify a fast exit, but we need to know
        // if the nonce was also valid from below us or not...
        return;
    }

    auto registered = is(entry);
    if(registered) {
        if(event.context() != entry.context()) {
            log_warn("invalid context for user {}", event.user());
            event.fail_message(SIP_FORBIDDEN);
            return;
        }


        log_notice("releasing user {}", event.user());
        auto& shared = reg_shared[entry.extension() - dialplan_min];
        shared.presence = ipc_reg::GONE;
        shared.activated = shared.expires = 0;
        active_map -= entry.extension();
        entry.drop();
    }
    else
        log_warn("cannot release {}", event.user());

    entry.receipt(event, 0);
}

void do_option(const args_t& args)
{
    const auto& event = args.event;
    auto to = std::string_view(event.to()->url->username);
    auto context = event.context();
    auto tid = event.tid();
    auto msg = exosip::msg_t(nullptr);
    auto code = SIP_OK;
    auto entry = registry();
    auto group = std::string{to};
    std::list<std::string> contacts;

    if(to == "system")
        contacts.emplace_front(sitefrom);
    else if(group_map.count(group)) {
        auto target = group_map.at(group);
        auto count = target.apply([&contacts](const auto& user) {
            contacts.emplace_back(user.local_uri());
            return true;
        });
        if(!count)
            code = SIP_TEMPORARILY_UNAVAILABLE;
    }
    else if(registry::is_extension(to)) {
        entry = mapped[util::get_value(to) - dialplan_min];
        if(is(entry))
            contacts.emplace_front(entry.local_uri());
        else
            code = SIP_TEMPORARILY_UNAVAILABLE;
    }
    else
        code = SIP_NOT_FOUND;

    exosip::guard_t lock(context);
    eXosip_options_build_answer(context, tid, code, &msg);
    if(!msg)
        return;

    for(const auto& contact : contacts)
        osip_message_set_header(msg, CONTACT, contact.c_str());

    if(is(entry))
        entry.allow(msg);
    else
        allow_local(msg);
    eXosip_options_send_answer(context, tid, code, msg);
}

void local_pub(const args_t& args)
{
    const auto& event = args.event;
    auto user = find(event.user());

    if(user.empty()) {
        event.fail_message(SIP_NOT_FOUND);
        return;
    }

    if(!is_authorized(user, event)) {
        log_error("failed to authorize {}", event.user());
        event.fail_message(SIP_FORBIDDEN);
        return;
    }

    auto to = event.to();
    auto tid = event.tid();
    auto type = event.content();
    auto body = content_view(event.body());
    auto target = "";
    auto entry = registry();
    auto publish = event.publish();
    auto expires = event.remains() / 1000;
    auto context = event.context();

    if(to && to->url && to->url->username) {
        target = to->url->username;
        entry = find(target);
    }

    if(type == "application/pidf+xml" && is(entry)) {
        auto shm = entry.shm();
        auto activity = (body.find("activities>") != std::string::npos);
        if(body.find("basic>open</") != std::string::npos)
            shm->presence = activity ? ipc_reg::AWAY : ipc_reg::HERE;
        else if(body.find("basic>closed</") != std::string::npos)
            shm->presence = activity ? ipc_reg::DND : ipc_reg::GONE;
    }

    debug(4, "publishing event {} with {} to {} until {} seconds", publish, type, target, expires);
    // TODO: check for presence & parse, send to exec::publish_event()
    eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
}

void local_msg(const args_t& args)
{
    const auto& event = args.event;
    auto user = find(event.user());
    auto system_user = group_map.at("system")[user.extension()];

    if(user.empty()) {
        event.fail_message(SIP_NOT_FOUND);
        return;
    }

    if(!is_authorized(user, event)) {
        log_error("failed to authorize {}", event.user());
        event.fail_message(SIP_FORBIDDEN);
        return;
    }

    registry_map target;
    auto context = event.context();
    auto body = event.body();
    auto type = event.content();
    auto text = content_view(body);
    auto tid = event.tid();
    auto to = std::string_view(event.to()->url->username);
    auto group = event.to()->url->username;
    auto from = user.from(event);
    auto subject = event.subject();

    debug(4, "message from \"{}\" for {}, text {}", from, to, text);

    if(system_user) {
        system_message(text, to, subject);
        exosip::guard_t lock(context);
        eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
        return;
    }

    if(to == "system") {
        while(!text.empty() && text.front() == ' ')
            text.remove_prefix(1);

        if(text.empty() || text.front() != '/' || !cli_command(user, text))
            cli_help(user);

        exosip::guard_t lock(context);
        eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
        return;
    }

    if(group_map.count(group))
        target = group_map.at(group);
    else if(registry::is_extension(to))
        target = coverage_map[util::get_value(to) - dialplan_min];
    else {
        event.fail_message(SIP_NOT_FOUND);
        return;
    }

    auto count = target.apply([&](const auto& entry) {
        if(!entry.allow_message())
            return false;

        auto route = entry.route_to();
        auto context = entry.context();
        auto to = entry.to(event);
        auto msg = message(from, to, route, subject, text, type, context);
        exec::deliver_local(msg);
        return true;
    });

    if(count)
        eXosip_message_send_answer(context, tid, SIP_OK, nullptr);
    else
        event.fail_message(SIP_TEMPORARILY_UNAVAILABLE);
}

void do_action(const args_t& args)
{
    const auto ipc = *args.ipc;
    auto ext = ipc.body.registry.ext;
    if(ext < dialplan_min || ext >dialplan_max) {
        proc::ipc_notify(ipc.pid, false);
        return;
    }

    auto id = std::to_string(ext);
    const char *key = ipc.body.registry.key;
    const char *value = ipc.body.registry.value;
    const auto& entry = mapped.at(ext - dialplan_min);
    auto shm = entry.shm();

    switch(ipc.type) {
    case IPC_ACTIVE:
        if(!is(entry)) {
            proc::ipc_notify(ipc.pid, false);
            return;
        }
        break;
    case IPC_CLEARS:
        key_storage.remove(id);
        key_updates = true;
        break;
    case IPC_ASSIGN:
        if(!key[0]) {
            proc::ipc_notify(ipc.pid, false);
            return;
        }
        key_storage[id][key] = value;
        key_updates = true;
        break;
    case IPC_PRESENCE:
        if(!is(entry)) {
            proc::ipc_notify(ipc.pid, false);
            return;
        }

        if(eq(value, "here"))
            shm->presence = ipc_reg::HERE;
        else if(eq(value, "away"))
            shm->presence = ipc_reg::AWAY;
        else if(eq(value, "dnd"))
            shm->presence = ipc_reg::DND;
        else if(eq(value, "gone"))
            shm->presence = ipc_reg::GONE;
        else {
            proc::ipc_notify(ipc.pid, false);
            return;
        }
        break;
    default:
        proc::ipc_notify(ipc.pid, false);
        return;
    }

    proc::ipc_notify(ipc.pid, true);
}

void system_msg(const args_t& args)
{
    auto& ipc = *args.ipc;
    auto from = system_from();
    if (!ipc.body.message.text[0]) {
        proc::ipc_notify(ipc.pid, true);
        return;
    }

    ipc.body.message.to[IPC_ID_MAX] = 0;
    ipc.body.message.text[IPC_SMS_MAX] = 0;
    ipc.body.message.subj[IPC_SUBJ_MAX] = 0;
    auto group = ipc.body.message.to;
    std::string_view to = ipc.body.message.to;
    std::string_view text = ipc.body.message.text;
    std::string_view subject = ipc.body.message.subj;
    registry_map target;

    debug(4, "system message for {}, text {}", to, text);
    if(to == "all")
        target = active_map;
    else if(group_map.count(group))
        target = group_map.at(group);
    else if(registry::is_extension(to))
        target = coverage_map[util::get_value(to) - dialplan_min];
    else {
        proc::ipc_notify(ipc.pid, false);
        return;
    }

    proc::ipc_notify(ipc.pid, target.apply([&](const auto& entry) {
        if(!entry.allow_message())
            return false;

        auto route = entry.route_to();
        auto context = entry.context();
        auto to = entry.origin_uri();
        auto msg = message(from, to, route, subject, text, "text/plain", context);
        exec::deliver_local(msg);
        return true;
    }) > 0);
}

void local_call(const args_t& args)
{
    const auto& event = args.event;
    auto user = find(event.user());

    if(!is_authorized(user, event)) {
        log_error("failed to authorize {}", event.user());
        event.fail_call(SIP_FORBIDDEN);
        return;
    }

    registry_map invites, target, delayed;
    auto to = std::string_view(event.to()->url->username);
    auto from = user.local_uri();
    auto group = event.to()->url->username;
    auto subject = event.subject();
    auto display = user.contact().display();
    //auto body = event.body();
    //auto sdp = content_view(body);
    auto busy = false;
    auto up = user.shm();

    // Check pre-conditions...
    if(group_map.count(group))
        invites = group_map.at(group);
    else if(registry::is_extension(to)) {
        auto map = util::get_value(to) - dialplan_min;
        invites = coverage_map[map];
        delayed = delayed_map[map];
    }

    // If nothing to invite, bad to...
    if(invites.empty()) {
        event.fail_call(SIP_NOT_FOUND);
        return;
    }

    // Make sure we have capacity...
    if(++up->count > up->lines) {
        --up->count;
        event.fail_call(SIP_BUSY_HERE);
        return;
    }

    debug(4, "call from \"{}\" for {}, subject {}", display, to, subject);

    // Attempt target calling...
    invites.apply([&](const auto& entry) {
        if(target[entry.extension()])
            return false;

        if(!entry.allow_invite())
            return false;

        // maybe forward busy support...etc...
        auto ep = entry.shm();
        if(++ep->count <=  ep->lines) {
            target += entry.extension();
            return true;
        }
        busy = true;
        --ep->count;
        return false;
    });

    // If invite targets, define for start ringing...
    if(!target.empty() || !delayed.empty()) {

        auto call_id = next_callid();
        if(call_id) {
            exec::create_call(call_id, event, user);
            target.apply([&call_id](const auto& to) {
                exec::invite_call(call_id, to);
                return true;
            });
            if(!delayed.empty()) {
                delayed ^= target;
                delayed.apply([&call_id](const auto& entry) {
                    if(!entry.allow_coverage())
                        return false;
                    exec::invite_delayed(call_id, entry);
                    return true;
                });
            }
            exec::start_call(call_id, event, user);
            return;
        }

        // if no call, clean up ring requests...
        target.apply([](const auto& entry) {
            --entry.shm()->count;
            return true;
        });
        debug(4, "call busy everywhere for {}", display);
        event.fail_call(SIP_BUSY_EVRYWHERE);
        return;
    }

    // If no targets or no call, clean up invite and exit...
    --up->count;
    debug(4, "call not possible for {}", display);
    if(busy)
        event.fail_call(SIP_BUSY_HERE);
    else
        event.fail_call(SIP_TEMPORARILY_UNAVAILABLE);
}
} // end anon namespace

registry::registry(const ext_t& key) :
ptr_(std::make_shared<data_t>())
{
    ptr_->extension = key;
    ptr_->count = 0;
    ptr_->contact.user_ = std::to_string(key);
    ptr_->contact.host_ = hostname;
    ptr_->contact.port_ = hostport;
    ptr_->contact.display_ = "Ext. " + ptr_->contact.user_;
    ptr_->welcome = true;

    shm_ = &reg_shared[extension() - dialplan_min];
    shm_->count = 0;
    shm_->lines = 0;
}

auto registry::is_extension(std::string_view text) -> bool
{
    if(text.empty())
        return false;

    auto value = 0U;
    for(auto ch : text) {
        if(!isdigit(ch))
            return false;
        value = (value * 10) + (ch - '0');
    }

    return !(value < dialplan_min || value > dialplan_max);
}

auto registry::local_uri() const -> std::string
{
    return ptr_ && context_ ? print_format("{}:{}", ptr_->schema, contact().id()) : "";
}

auto registry::local_alias(std::string_view to) const -> std::string
{
    return ptr_ && context_ ? print_format("{}:{}@{}", ptr_->schema, to, contact().addr()) : "";
}

auto registry::local_to() const -> std::string
{
    return ptr_ && context_ ? print_format("{}:{}", ptr_->schema, contact().addr()) : "";
}

auto registry::origin_uri() const -> std::string
{
    return ptr_ && context_ ? print_format("{}:{}", ptr_->schema, origin().id()) : "";
}

auto registry::route_to() const -> std::string
{
    return ptr_ && context_ ? print_format("{}:{}", ptr_->schema, origin().addr()) : "";
}

void registry::keys(const keyfile::keys& pairs)
{
    if(ptr_) {
        std::string sha256, md5, secret;
        auto user = std::to_string(extension());

        try {
            shm_->lines = std::stoi(pairs.at("lines"));
        }
        catch(...) {
            shm_->lines = 1;
        }

        try {
            secret = pairs.at("secret");
        }
        catch(...) {
            try {
                secret = pairs.at("password");
            }
            catch(...) {}
        }

        if(!secret.empty()) {
            sha256 = util::sha256_secret(user, realm, secret);
            md5 = util::md5_secret(user, realm, secret);
        }

        std::lock_guard lock(ptr_->mutex);
        ptr_->keys = pairs;
        if(chk_secret(user, sha256, 64))
            ptr_->keys["sha256"] = sha256;
        if(chk_secret(user, md5, 32))
            ptr_->keys["md5"] = md5;

        try {
            ptr_->contact.display_ = pairs.at("display");
        }
        catch(...) {
            ptr_->contact.display_ = "Ext. " + ptr_->contact.user_;
        }
    }
}

auto registry::keys() const -> keyfile::keys
{
    keyfile::keys tmp;

    if(ptr_) {
        std::lock_guard lock(ptr_->mutex);
        tmp.insert(ptr_->keys.begin(), ptr_->keys.end());
    }
    return tmp;
}

auto registry::contact() const -> contact_t
{
    if(ptr_) {
        std::lock_guard lock(ptr_->mutex);
        return ptr_->contact;
    }
    return {};
}

auto registry::to(const event_t& event) const -> std::string
{
    auto id = std::string_view(event.to()->url->username);
    auto display = event.to()->displayname;
    if(display && *display)
        return print_format("\"{}\" <{}>", display, local_alias(id));
    return print_format("<{}>", local_alias(id));
}

auto registry::from([[maybe_unused]] const event_t& event) const -> std::string
{
    if(event) {
        auto id = std::string_view(event.from()->url->username);
        auto group = std::string(event.from()->url->username);
        auto display = event.from()->displayname;
        group[0] = static_cast<char>(std::toupper(group[0]));
        group += " Group";
        if(!display || !*display)
            display = const_cast<char *>(group.c_str());
        if(!registry::is_extension(id))
            return print_format("\"{}\" <{}>", display, local_alias(id));
    }
    auto display = contact().display();
    auto phone = ";user=phone";
    if(display.empty())
        return print_format("<{}{}>", local_uri(), phone);
    return print_format("\"{}\" <{}{}>", display, local_uri(), phone);
}

auto registry::origin() const -> contact_t
{
    if(ptr_) {
        std::lock_guard lock(ptr_->mutex);
        return ptr_->origin;
    }
    return {};
}

registry::operator bool() const
{
    if(!context_)
        return false;
    time_t now{};
    time(&now);
    return expires() > now;
}

auto registry::operator!() const -> bool
{
    if(!context_)
        return true;
    time_t now{};
    time(&now);
    return expires() <= now;
}

void registry::receipt(const event_t& event, time_t period) // NOLINT
{
    auto context = event.context();
    auto tid = event.tid();
    auto msg = exosip::msg_t(nullptr);
    auto expires = std::to_string(period);

    exosip::guard_t lock(context);
    eXosip_message_build_answer(context, tid, SIP_OK, &msg);
    osip_message_set_header(msg, "Expires", expires.c_str());
    eXosip_message_send_answer(context, tid, SIP_OK, msg);
}

auto registry::refresh(const event_t& event, time_t period) -> bool
{
    if(!ptr_ || !event.request())
        return false;

    time_t now{};
    time(&now);

    auto registered = ptr_->expires > now;
    if(!registered) {
        auto request = event.request();
        context_ = event.context();
        ptr_->schema = event.schema();
        ptr_->agent = event.agent();

        auto allows = request->allows;
        if(osip_list_size(&allows)) {
            ptr_->invite = ptr_->message = false;
            auto pos = 0;
            while(osip_list_eol(&allows, pos) == 0) {
                auto allow = std::string_view(static_cast<osip_allow_t *>(osip_list_get(&allows, pos++))->value);
                if(allow == "INVITE")
                    ptr_->invite = true;
                else if(allow == "MESSAGE")
                    ptr_->message = true;
            }
        }
        else {
            ptr_->invite = ptr_->message = true;
        }
    }
    else if(context_ != event.context())
        return true;

    if(!event.count().empty())
        ptr_->count = util::get_count(event.count());

    ptr_->nonce = nonce_t(event.nonce());
    if(period) {
        now += period;
        ptr_->expires = now;
        // update origin in lock
        std::lock_guard lock(ptr_->mutex);
        ptr_->origin.from(event);
    }
    else
        drop();

    return registered;
}

auto registry::verify(const event_t& event) -> bool
{
    if(!ptr_) {
        debug(5, "invalid user to verify");
        return false;
    }

    auto count = event.count();
    auto client = nonce_t(event.client());

    // if no nonce count, and none registered this session, okay...
    if(count.empty() && !ptr_->count)
        return true;

    debug(5, "client nounce {}", count);
    auto value = util::get_count(count);
    if(value > ptr_->count) {
        ptr_->client = client.hex();
        ptr_->count = value;
        return true;
    }

    if(value < ptr_->count) {
        debug(5,"failed counters: {} != {}", value, ptr_->count);
        return false;
    }

    if(client.hex() != ptr_->client) {
        ptr_->client = client.hex();
        return true;
    }

    debug(5, "failed client:{} == {}", client.hex(), ptr_->count);
    return false;
}

void registry::drop()
{
    if(!ptr_ || !ptr_->expires)
        return;

    exec::hangup_user(*this, reason_t(SIP_REQUEST_TERMINATED, "user dropped"));
    debug(5, "removing user {} from {}", extension(), remote());
    clients.erase(remote());
    ptr_->expires = 0;
    ptr_->count = 0;
    notify();
}

void registry::allow(exosip::msg_t msg)
{
    auto allows = std::string("OPTIONS,REGISTER,INFO");
    auto accepts = std::string("text/plain");

    if(ptr_ && ptr_->message) {
        allows += ",MESSAGE";
        osip_message_set_header(msg, ACCEPT_ENCODING, "text/plain");
    }
    if(ptr_ && ptr_->invite) {
        allows += ",INVITE,BYE,CANCEL,ACK";
        accepts += ",application/sdp,application/dtmf-relay";
    }
    osip_message_set_header(msg, ALLOW, allows.c_str());
    osip_message_set_header(msg, ACCEPT, accepts.c_str());
}

void server::change_level(unsigned level)
{
    reg_system->level = logging = level;
    log_notice("logging level changed to {}", logging);
}

void server::system_message(std::string_view text, std::string_view to, std::string_view subject)
{
    struct ipc_msg msg{};
    memset(&msg, 0, sizeof(msg));
    msg.pid = 0;
    msg.type = IPC_MESSAGE;
    util::set_string(msg.body.message.to, sizeof(msg.body.message.to), to);
    util::set_string(msg.body.message.text, sizeof(msg.body.message.text), text);
    if(!subject.empty())
        util::set_string(msg.body.message.subj, sizeof(msg.body.message.subj), subject);
    exec::message_system(msg);
}

void server::exec_registry(const keyfile& config)
{
    using namespace std::chrono;

    auto args = args_t(config);
    auto shm = shm_open(SERVER_IPCPATH ".registry", O_RDWR | O_CREAT, 0660);
    if(shm < 0 || ftruncate(shm, mem_size) < 0)
        log_fail(97, "failed to allocate registry");
    auto mem = mmap(nullptr, mem_size, PROT_WRITE, MAP_SHARED, shm, 0);
    if(mem == MAP_FAILED)    // NOLINT
        log_fail(97, "failled to map registry");

    memset(mem, 0, mem_size);
    reg_shared = static_cast<ipc_reg *>(mem);
    reg_system = reinterpret_cast<ipc_sys *>(&reg_shared[reg_entries]);

    key_storage.load("registry.keys");
    reg_system->version = IPC_VERSION;
    time(&reg_system->started);
    reg_system->first = dialplan_min;
    reg_system->level = logging;
    reg_system->pid = getpid();
    do_reload(args);

    log_info("starting registry thread");
    while(!stopping) {
        std::unique_lock<std::mutex> lock(msglock);
        if(!msgs.empty()) {
            auto copy = std::move(msgs.front());
            auto [actor, args] = copy;
            msgs.pop_front();
            lock.unlock();
            try {
                actor(args);
            }
            catch(const std::exception& e) {
                log_error("registry actor error: {}", e.what());
            }
            continue;
        }

        time_t latest = 0;
        if(timer) {
            lock.unlock();
            latest = rescan();
            lock.lock();
            if(!msgs.empty())
                continue;
            timer = false;
        }

        if(latest)
            timer = (cvar.wait_until(lock, system_clock::from_time_t(latest)) == std::cv_status::timeout);
        else
            cvar.wait(lock);
    }
    log_info("stopping registry thread");
    sync_storage();
    munmap(reg_shared, mem_size);
    ::close(shm);
}

contact_t::contact_t(const osip_uri_t *uri)
{
    if(!uri || !uri->host || uri->host[0] == 0)
        return;

    host_ = uri->host;
    if(uri->scheme && uri->scheme == "sips"s)
        port_ = 5061;
    else
        port_ = 5060;

    if(uri->port && uri->port[0] != 0)
        port_ = util::get_port(uri->port);

    if(uri->username)
        user_ = uri->username;
}

void contact_t::from(const event_t& event)
{
    auto clist = event.request()->contacts;
    auto pos = 0;
    auto port = uint16_t(5060);

    osip_uri_t *url{nullptr};
    while(osip_list_eol(&clist, pos) == 0) {
        auto c = static_cast<osip_contact_t *>(osip_list_get(&clist, pos++));
        if(c && c->url && c->url->host) {
            url = c->url;
            // TODO: is ipv4 to be treated as a subset??
            /* if(ipv6 && net::is_ipv4(c->url->host))
                continue; */
            if(!ipv6 && net::is_ipv6(c->url->host))
                continue;
            break;
        }
    }

    if(url && url->username)
        user_ = url->username;
    if(url && url->host)
        host_ = url->host;
    if(url && url->port && url->port[0])
        port = port_ = util::get_port(url->port);

    auto vlist = event.request()->vias;
    pos = 0;
    while(osip_list_eol(&vlist, pos) == 0) {
        auto via = static_cast<osip_via_t *>(osip_list_get(&vlist, pos++));
        auto rport = uint16_t(0);
        if(!via->host || !via->host[0])
            continue;
        if(via->port && via->port[0])
            port = util::get_port(via->port);
        osip_uri_param_t *param{nullptr};
        osip_via_param_get_byname(via, const_cast<char *>("rport"), &param);
        if(param && param->gvalue && param->gvalue[0])
            rport = util::get_port(param->gvalue);

        // detect top nat...
        if(rport) {
            param = nullptr;
            osip_via_param_get_byname(via, const_cast<char *>("received"), &param);
            if(param && param->gvalue && param->gvalue[0]) {
                host_ = param->gvalue;
                port_ = rport;
                break;
            }
        }
        else {
            host_ = via->host;
            port_ = port;
        }
    }

    if(host_[0] != '[' && host_.find(':') != std::string::npos)
        host_ = "[" + host_ + "]";
}

void exec::stop_registry()
{
    dispatch(&do_shutdown);
}

void exec::reload_registry(const keyfile& config)
{
    dispatch(&do_reload, config);
}

void exec::message_system(const ipc_msg& msg)
{
    dispatch(&system_msg, std::make_shared<ipc_msg>(msg));
}

void exec::ipc_action(const ipc_msg& ipc)
{
    dispatch(&do_action, std::make_shared<ipc_msg>(ipc));
}

void exec::release_user(const event_t& event)
{
    dispatch(&do_release, event);
}

void exec::refresh_user(const event_t& event)
{
    dispatch(&do_refresh, event);
}

void exec::option_user(const event_t& event)
{
    dispatch(&do_option, event);
}

void exec::publish_user(const event_t& event)
{
    dispatch(&local_pub, event);
}

void exec::message_user(const event_t& event)
{
    dispatch(&local_msg, event);
}

void exec::invite_user(const event_t& event)
{
    dispatch(&local_call, event);
}

