/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "ipc.hpp"

#include <print.hpp>
#include <args.hpp>

#include <csignal>

#include <unistd.h>
#include <fcntl.h>
#include <mqueue.h>
#include <sys/mman.h>
#include <arpa/inet.h>
#include <netdb.h>

namespace {
args::option helpflag('h',  "--help",       "display this list");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option to(0,  "--to", "to user or group", "id");
args::option reload('r',    "--reload", "reload server");
args::option subject('s',   "--subject", "message topic", "text");
args::option timeout('t',   "--timeout", "reply wait", "seconds");
args::option althelp('?',   nullptr, nullptr);

class ipc_error final : public std::exception
{
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

auto scount(const char *cp, size_t max = 256)
{
    size_t count = 0;
    while(cp && *cp && count < max) {
        ++count;
        ++cp;
    }
    if(count >= max)
        count = 0;

    return count;
}

auto scopy(char *cp, size_t max, std::string_view from) -> size_t
{
    auto count = from.size();
    auto dp = from.data();
    if(count >= max)
        count = max - 1;

    max = count;
    while(max--)
        *(cp++) = *(dp++);

    *cp = 0;
    return count;
}
} // end namespace


auto main(int argc, const char **argv) -> int   // NOLINT
{
    auto exit_code = 0;
    try {
        auto wait_time = 5;
        auto id = "all";
        auto topic = "";
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count != 1) {
            args::help({"pbx-message [options] \"text\""}, {
                "pbx-message " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Send system message to pbx users"
            });
            return 0;
        }
        auto text = args::list()[0];

        if(is(to))
            id = *to;

        if(!scount(id, IPC_ID_MAX))
            throw ipc_error("user id too long", 10);

        if(text.empty() || text.size() > IPC_SMS_MAX)
            throw ipc_error("text size is invalid", 11);

        if(is(subject)) {
            topic = *subject;
            if(!scount(topic, IPC_SUBJ_MAX))
                throw ipc_error("subject is too long", 12);
        }

        signal(SIGUSR2, []([[maybe_unused]]int signo){
            throw ipc_error("message or to not accepted", 1);
        });

        signal(SIGUSR1, []([[maybe_unused]]int signo){
            ::exit(0);
        });

        ipc_msg msg{.type = IPC_MESSAGE, .pid = getpid(), .ver = IPC_VERSION};
        if(is(reload))
            msg.type = IPC_RELOAD;
        scopy(msg.body.message.to, IPC_ID_MAX, id);
        scopy(msg.body.message.text, IPC_SMS_MAX, text);
        scopy(msg.body.message.subj, IPC_SUBJ_MAX, topic);
        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq < 0)
            throw ipc_error("mqueue failed", 4);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);
        ::sleep(wait_time);
        throw ipc_error("timeout", 2);
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-message [options] \"text\"\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}
