/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "ipc.hpp"

#include <keyfile.hpp>
#include <print.hpp>
#include <strings.hpp>
#include <args.hpp>

#include <csignal>

#include <unistd.h>
#include <fcntl.h>
#include <mqueue.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>

namespace {
args::option clearing('c',  "--clear",      "clear users registry keys");
args::option helpflag('h',  "--help",       "display this list");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option timeout('t',   "--timeout",    "reply wait", "seconds");
args::option althelp('?',   nullptr, nullptr);

std::string keyname;
auto count = 0U;
auto extension = 0;

class ipc_error final : public std::exception
{
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

auto get_value(std::string_view text) -> int
{
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

auto set_string(char *cp, size_t max, std::string_view view) -> size_t
{
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = *(dp++);

    *cp = 0;
    return size;
}

auto set_lower(char *cp, size_t max, std::string_view view) -> size_t
{
    assert(cp != nullptr);

    auto size = view.size();
    auto dp = view.data();
    if(size >= max)
        size = max - 1;

    max = size;
    while(max--)
        *(cp++) = static_cast<char>(tolower(*(dp++)));

    *cp = 0;
    return size;
}
} // end namespace

auto main(int argc, const char **argv) -> int   // NOLINT
{
    auto exit_code = 0;
    try {
        count = args::parse(argc, argv);
        auto wait_time = 5;
        auto list = args::list();
        if(is(helpflag) || is(althelp) || count < 1 || count > 3) {
            args::help({"pbx-registry [options] extension [key [value]]"}, {
                "pbx-registry " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Examine registry, see and modify user keys"
            });
            return 0;
        }

        if(::chdir(SERVER_PREFIX_PATH))
            throw std::runtime_error(print_format("cannot_access {}", SERVER_PREFIX_PATH));

        extension = get_value(list[0]);

        signal(SIGUSR2, []([[maybe_unused]]int signo){
            throw ipc_error("offline or invalid", 1);
        });

        signal(SIGUSR1, []([[maybe_unused]]int signo){
            if(!is(quiet)) {
                if(is(clearing))
                    print("user {} keys cleared\n", extension);
                else if(count == 3)
                    print("user {} key set\n", extension);
                else if(count == 1)
                    print("user {} exists\n", extension);
            }
            if(count == 2) {
                keyfile keys({SERVER_CONFIG_FILE, "dynamic.conf", "state.conf", "custom.conf", "registry.keys"});
                auto user = keys[std::to_string(extension)];
                auto common = keys["common"];
                auto value = std::string();
                try {
                    value = user.at(keyname);
                }
                catch(...) {
                    try {
                        value = common.at(keyname);
                    }
                    catch(...) {}
                }
                print("{}\n", value);
            }
            ::exit(0);
        });

        if(is(timeout))
            wait_time = std::stoi(*timeout);

        ipc_msg msg{.type = IPC_ACTIVE, .pid = getpid(), .ver = IPC_VERSION};
        msg.body.registry.ext = extension;
        if(count > 1)
            keyname = lower_case(std::string{list[1]});
        if(is(clearing))
            msg.type = IPC_CLEARS;
        else if(count == 3) {
            set_lower(msg.body.registry.key, IPC_KEY_SIZE, list[1]);
            set_string(msg.body.registry.value, IPC_SET_SIZE, list[2]);
            msg.type = IPC_ASSIGN;
        }
        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq < 0)
            throw ipc_error("mqueue failed", 4);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);
        ::sleep(wait_time);
        throw ipc_error("timeout", 2);
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-registry [options] extensions [key [value]]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}
