/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "ipc.hpp"

#include <print.hpp>
#include <args.hpp>

#include <csignal>

#include <unistd.h>
#include <fcntl.h>
#include <mqueue.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>

namespace {
args::option helpflag('h',  "--help",       "display this list");
args::option quiet('q',     "--quiet",      "suppress error output");
args::option timeout('t',   "--timeout",    "reply wait", "seconds");
args::option althelp('?',   nullptr, nullptr);

class ipc_error final : public std::exception
{
public:
    explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

auto is_file(const std::string& path) -> bool
{
    struct stat ino{};
    if(::lstat(path.c_str(), &ino) < 0)
        return false;
    return S_ISREG(ino.st_mode);
}
} // end namespace

auto main(int argc, const char **argv) -> int   // NOLINT
{
    auto exit_code = 0;
    try {
        auto wait_time = 5;
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count > 1) {
            args::help({"pbx-runstate [new-state]"}, {
                "pbx-runstate " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Change pbx running mode"
            });
            return 0;
        }

        if(::chdir(SERVER_PREFIX_PATH))
            throw std::runtime_error(print_format("cannot_access {}", SERVER_PREFIX_PATH));

        char symbuf[22];
        if(!count) {
            auto size = readlink("state.conf", symbuf, sizeof(symbuf) - 1); // FlawFinder: ignore
            if(size > 0) {
                symbuf[size] = 0;
                if(symbuf[0] == '.' || strchr(symbuf, '/'))
                    symbuf[0] = 0;
                else {
                    auto ep = strrchr(symbuf, '.');
                    if(ep && is_file(symbuf))
                        *ep = 0;
                    else
                        symbuf[0] = 0;
                }
            }
            else
                symbuf[0] = 0;

            if(!symbuf[0])
                print("default\n");
            else
                print("{}\n", symbuf);

            ::exit(0);
        }

        auto state = std::string{args::list()[0]};
        if(state == "default") {
            ::remove("state.conf");
        }
        else {
            snprintf(symbuf, sizeof(symbuf), "%s.state", state.data());
            if(!is_file(symbuf))
                throw std::runtime_error(print_format("{}: unknown state", symbuf));
            ::remove("state.conf");
            ::symlink(symbuf, "state.conf");
        }

        signal(SIGUSR2, []([[maybe_unused]]int signo){
            throw ipc_error("reload failed", 1);
        });

        signal(SIGUSR1, []([[maybe_unused]]int signo){
            ::exit(0);
        });

        if(is(timeout))
            wait_time = std::stoi(*timeout);

        ipc_msg msg{.type = IPC_RELOAD, .pid = getpid(), .ver = IPC_VERSION};
        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq < 0)
            ::exit(0);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);
        ::sleep(wait_time);
        throw ipc_error("timeout", 2);
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-runstate [new-state]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}
