/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "ipc.hpp"

#include <print.hpp>
#include <args.hpp>

#include <csignal>

#include <unistd.h>
#include <fcntl.h>
#include <mqueue.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>

namespace {
args::option callflag('c',  "--calls",      "show active calls");
args::option helpflag('h',  "--help",       "display this list");
args::option loglevel('l',  "--level",      "set log level", "0-9");
args::option quiet('q',     "--quiet",      "suppress all output");
args::option registry('r',  "--registry",   "show registry status");
args::option reload(0,      "--reload",     "reload server config");
args::option timeout('t',   "--timeout",    "reply wait", "seconds");
args::option althelp('?',   nullptr, nullptr);

class ipc_error final : public std::exception
{
public:
    inline explicit ipc_error(const char *text, int code) noexcept : msg(text), err(code) {}

    auto what() const noexcept -> const char * override {
        return msg;
    }

    auto exit() const noexcept {
        return err;
    }

private:
    const char *msg;
    int err;
};

auto scount(const char *cp, size_t max = 256)
{
    size_t count = 0;
    while(cp && *cp && count < max) {
        ++count;
        ++cp;
    }
    return count;
}

} // end namespace

auto main(int argc, const char **argv) -> int   // NOLINT
{
    auto exit_code = 0;
    try {
        auto wait_time = 5;
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || count) {
            args::help({"pbx-status [options]"}, {
                "pbx-status " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com",
                "Send command to server ipc"
            });
            return 0;
        }

        signal(SIGUSR2, []([[maybe_unused]]int signo){
            throw ipc_error("offline", 1);
        });

        signal(SIGUSR1, []([[maybe_unused]]int signo){
            auto calls = 0U;
            auto count = 0U;
            auto busy = 0U;
            auto rmax = 80U;

            auto cshm = shm_open(SERVER_IPCPATH ".calls", O_RDONLY, 0660);
            if(cshm < 0)
                throw ipc_error("mapping failed", 4);
            struct stat cino{};
            if(fstat(cshm, &cino) < 0)
                throw ipc_error("stat failed", 4);
            auto cmax = cino.st_size / sizeof(ipc_call);
            auto cmem = mmap(nullptr, sizeof(ipc_call) * cmax, PROT_READ, MAP_SHARED, cshm, 0);
            if(cmem == MAP_FAILED)   // NOLINT
                throw ipc_error("mapping failed", 4);

            auto rshm = shm_open(SERVER_IPCPATH ".registry", O_RDONLY, 0660);
            if(rshm < 0)
                throw ipc_error("mapping failed", 4);
            struct stat rino{};
            if(fstat(rshm, &rino) < 0)
                throw ipc_error("stat failed", 4);
            rmax = (rino.st_size - sizeof(ipc_sys)) / sizeof(ipc_reg);
            auto rmem = mmap(nullptr, sizeof(ipc_reg) * rmax + sizeof(ipc_sys), PROT_READ, MAP_SHARED, rshm, 0);
            if(rmem == MAP_FAILED)   // NOLINT
                throw ipc_error("mapping failed", 4);

            auto call = static_cast<ipc_call *>(cmem);
            auto reg = static_cast<ipc_reg*>(rmem);
            auto sys = reinterpret_cast<ipc_sys *>(&reg[rmax]);
            auto trunk = &sys->trunk;
            time_t now{};
            time(&now);
            now -= sys->started;

            for(auto i = 0U; i < cmax; ++i) {
                const auto& entry = call[i];
                if(entry.id > 0)
                    ++calls;
                if(is(callflag) && entry.id > 0) {
                    auto status = "unknown";
                    switch (entry.state) {
                    case ipc_call::INVITING:
                        status = "inviting";
                        break;
                    case ipc_call::RINGING:
                        status = "ringing";
                        break;
                    case ipc_call::CLOSING:
                        status = "closing";
                        break;
                    case ipc_call::CONNECTED:
                        status = "active";
                        break;
                    }

                    auto type = "unknown";
                    switch(entry.type) {
                    case ipc_call::LOCAL:
                        type = "local";
                        break;
                    case ipc_call::INBOUND:
                        type = "inbound";
                        break;
                    case ipc_call::OUTGOING:
                        type = "outgoing";
                        break;
                    case ipc_call::NONE:
                        type = "none";
                        break;
                    }

                    time_t duration{};
                    time(&duration);
                    duration -= entry.created;

                    print("0x{:08x} {} {}, {} participant(s), {}s duration\n", entry.id, status, type, entry.ringing ? entry.ringing : entry.active, duration);
                }
            }

            for(auto i = 0U; i < rmax; ++i) {
                const auto& entry = reg[i];
                time_t expires{};
                time(&expires);
                if(!entry.activated || expires > entry.expires)
                    continue;

                auto presence = "here";
                if(entry.count >= entry.lines)
                    presence = "busy";
                else if(entry.count)
                    presence = "call";
                switch(entry.presence) {
                case ipc_reg::GONE:
                    presence = "gone";
                    break;
                case ipc_reg::DND:
                    presence = "dnd ";
                    break;
                case ipc_reg::AWAY:
                    presence = "away";
                    break;
                default:
                    break;
                }
                if(is(registry)) {
                    auto addr = reinterpret_cast<const struct sockaddr *>(&entry.address);
                    char host[128];
                    uint16_t port = 0;
                    memset(host, 0, sizeof(host));
                    switch(addr->sa_family) {
                    case AF_INET:
                        getnameinfo(addr, sizeof(struct sockaddr_in), host, sizeof(host), nullptr, 0, NI_NUMERICHOST);
                        port = ntohs(reinterpret_cast<const struct sockaddr_in*>(addr)->sin_port);
                        break;
                    case AF_INET6:
                        getnameinfo(addr, sizeof(struct sockaddr_in6), host + 1, sizeof(host) - 2, nullptr, 0, NI_NUMERICHOST);
                        host[0] = '[';
                        host[scount(host)] = ']';
                        port = ntohs(reinterpret_cast<const struct sockaddr_in6*>(addr)->sin6_port);
                        break;
                    default:
                        snprintf(host, sizeof(host), "unknown");
                    }
                    auto endpoint = print_format("{}:{}", host, port);
                    auto used = entry.count.load();
                    if(used > entry.lines)
                        used = entry.lines;
                    print("{:<4}{:32} {:4} {:8}s {}/{}\n", i + sys->first, endpoint, presence, entry.expires - expires, used, entry.lines);
                }
                if(entry.count)
                    ++busy;
                ++count;
            }
            if(!is(quiet)) {
                auto remotes = trunk->count.load();
                if(remotes > trunk->lines)
                    remotes = trunk->lines;
                print("online={}/{} busy={} calls={}/{} remotes={}/{} uptime={}m log={}\n", count, rmax, busy, calls, cmax, remotes, trunk->lines, now / 60, sys->level);
            }
            ::exit(0);
        });

        if(is(timeout))
            wait_time = std::stoi(*timeout);

        ipc_msg msg{.type = IPC_STATUS, .pid = getpid(), .ver = IPC_VERSION};
        if(is(loglevel)) {
            msg.type = IPC_LEVEL;
            msg.body.level = std::stoi(*loglevel);
        }
        else if(is(reload)) {
            msg.type = IPC_RELOAD;
            msg.body.message.text[0] = 0;
        }

        auto mq = mq_open(SERVER_IPCPATH, O_WRONLY, 0660, nullptr);
        if(mq < 0)
            throw ipc_error("mqueue failed", 4);
        if(mq_send(mq, reinterpret_cast<const char *>(&msg), sizeof(msg), 0) < 0)
            throw ipc_error("send failed", 4);
        ::sleep(wait_time);
        throw ipc_error("timeout", 2);
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: pbx-status [options]\nerror: {}\n", e.what());
    }
    catch(const ipc_error& e) {
        exit_code = e.exit();
        if(!is(quiet))
            die(exit_code, "error: {}\n", e.what());
    }
    catch(const std::exception& e) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        if(is(quiet))
            exit_code = -1;
        else
            die(-1, "error: unknown error\n");
    }
    return exit_code;
}
