/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef EXOSIP_HPP_
#define EXOSIP_HPP_

#include <memory>
#include <osipparser2/osip_port.h>
#include <string>
#include <cstring>
#include <cstdlib>
#include <eXosip2/eXosip.h>
#include <sys/types.h>
#ifdef WIN32
#include <WinSock2.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#endif

namespace exosip {
enum class listen_type {
    udp, tcp, tls, dtls
};

enum class family_type : int {
    ipv4 = AF_INET, ipv6 = AF_INET6,
};

enum class tls_error: int {
    none = TLS_OK,
    random = TLS_ERR_NO_RAND,
    dhparam = TLS_ERR_NO_DH_PARAM,
    password = TLS_ERR_NO_PW,
    root = TLS_ERR_NO_ROOT_CA,
    cert = TLS_ERR_MISSING_AUTH_PART,
};

using msg_t = osip_message_t*;
using body_t = osip_body_t*;
using header_t = osip_header_t*;

inline void release(void *p) {
    if(p != nullptr) {
        if(osip_free_func)
            osip_free_func(p);
        else {
            ::free(p); // NOLINT
        }
    }
}

class certificate_t final {
public:
    inline certificate_t() { // NOLINT
        memset(&tls_, 0, sizeof(tls_));
    }

    inline auto operator*() const {
        return &tls_;
    }

    inline auto set_certificate(const std::string& crtfile) {
        snprintf(tls_.client.cert, sizeof(tls_.client.cert),
            "%s", crtfile.c_str());
    }

    inline auto set_authority(const std::string& cafile) {
        snprintf(tls_.root_ca_cert, sizeof(tls_.root_ca_cert),
            "%s", cafile.c_str());
    }

    inline auto set_keyfile(const std::string& keyfile) {
        snprintf(tls_.client.priv_key, sizeof(tls_.client.priv_key),
            "%s", keyfile.c_str());
    }

    inline auto set_password(const std::string& password) {
        snprintf(tls_.client.priv_key_pw, sizeof(tls_.client.priv_key_pw),
            "%s", password.c_str());
    }

private:
    eXosip_tls_ctx_t tls_;
};

class context_t final {
public:
    inline explicit operator eXosip_t *() const { return ctx_;}
    inline explicit operator bool() const { return ctx_ != nullptr;}
    inline auto operator!() const { return ctx_ == nullptr;}
    inline auto operator*() const { return ctx_;}

    class guard final {
    public:
        inline explicit guard(const context_t& from) : context_(*from) {
            eXosip_lock(context_);
        }

        inline explicit guard(eXosip_t *from) : context_(from) {
            eXosip_lock(context_);
        }

        inline ~guard() {
            eXosip_unlock(context_);
        }

    private:
        eXosip_t *context_;
    };

    inline explicit context_t(const char *id) noexcept : port_(0), proto_(0), family_(AF_INET), schema_("sip"), ctx_(eXosip_malloc()) {
        eXosip_init(ctx_);
        eXosip_set_user_agent(ctx_, id);
    }

    inline ~context_t() {
        eXosip_quit(ctx_);
        release(ctx_);
    }

    context_t() = delete;
    context_t(const context_t& from) = delete;
    auto operator=(const context_t& from) -> context_t& = delete;

    inline auto family() const {
        return family_;
    }

    inline auto schema() const {
        return schema_;
    }

    inline auto active() const {
        return proto_ != 0;
    }

    inline auto tcp() const {
        return proto_ == IPPROTO_TCP;
    }

    inline auto port() const {
        return proto_ ? port_ : 0;
    }

    inline auto close() noexcept {
        if(ctx_ && proto_) {
            eXosip_quit(ctx_);
            proto_ = 0;
        }
    }

    inline auto find_port(int port = 5062, int proto = IPPROTO_UDP) const {
        for(;;) {
            port = eXosip_find_free_port(ctx_, port, proto);
            if(port & 0x01)
                ++port;
            else
                break;
        }
        return port;
    }

    // set_option uses different types, so we create type specific functions

    inline auto reuse_tcp_port(bool enable) {
        int value = enable ? 1 : 0;
        eXosip_set_option(ctx_, EXOSIP_OPT_ENABLE_REUSE_TCP_PORT, &value);
    }

    inline auto use_ephemeral_port(bool enable) {
        int value = enable ? 1 : 0;
        eXosip_set_option(ctx_, EXOSIP_OPT_ENABLE_USE_EPHEMERAL_PORT, &value);
    }

    inline auto set_family(family_type family) {
        int value = family == family_type::ipv6 ? 1 : 0;
        family_ = value ? AF_INET6 : AF_INET;
        eXosip_set_option(ctx_, EXOSIP_OPT_ENABLE_IPV6, &value);
    }

    inline auto use_rport(bool enable) {
        int value = enable ? 1 : 0;
        eXosip_set_option(ctx_, EXOSIP_OPT_USE_RPORT, &value);
    }

    inline auto srv_lookup(bool enable) {   // enabled by default
        int value = enable ? 2 : 0;
        eXosip_set_option(ctx_, EXOSIP_OPT_DNS_CAPABILITIES, &value);
    }

    inline auto set_keep_alive(int msecs) { // common value 17000
        if(proto_ == IPPROTO_UDP) {
            eXosip_set_option(ctx_, EXOSIP_OPT_UDP_KEEP_ALIVE, &msecs);
            return true;
        }
        return false;
    }

    inline auto set_user_agent(const std::string& name) {
        eXosip_set_user_agent(ctx_, name.c_str());
    }

    inline auto set_dscp(int value = 26) {
        eXosip_set_option(ctx_, EXOSIP_OPT_SET_DSCP, &value);
    }

    inline auto set_gateway(const std::string& host) {
        if(family_ == AF_INET6)
            return eXosip_set_option(ctx_,
                EXOSIP_OPT_SET_IPV6_FOR_GATEWAY, host.c_str()) == 0;
        return eXosip_set_option(ctx_,
            EXOSIP_OPT_SET_IPV4_FOR_GATEWAY, host.c_str()) == 0;
    }

    inline auto set_verify(bool enable) {
        int value = enable ? 1 : 0;
        eXosip_set_option(ctx_, EXOSIP_OPT_SET_TLS_VERIFY_CERTIFICATE, &value);
    }

    inline auto set_certificate(const certificate_t& cert) {
        return static_cast<tls_error>(eXosip_set_option(ctx_,
            EXOSIP_OPT_SET_TLS_CERTIFICATES_INFO, *cert));
    }

    inline auto listen(const std::string& host = "*", listen_type type = listen_type::udp, uint16_t port = 5060) {
        auto addr = host.c_str();
        auto secure = (type == listen_type::dtls || type == listen_type::tls) ? 1 : 0;
        port_ = port;
        proto_ = (type == listen_type::udp || type == listen_type::dtls) ? IPPROTO_UDP : IPPROTO_TCP;
        if(addr[0] == 0 || addr[0] == '*' || host == "::*" || host == "::")
            addr = nullptr;
        schema_ = secure ? "sips" : "sip";
        if(!eXosip_listen_addr(ctx_, proto_, addr, port,  family_, secure))
            return true;
        proto_ = 0;
        return false;
    }

    inline auto automatic() {
        guard lock(ctx_);
        eXosip_automatic_action(ctx_);
    }

private:
    uint16_t port_;
    int proto_;
    int family_;
    std::string schema_;
    eXosip_t *ctx_;
};

using guard_t = context_t::guard;

class event_t {
public:
    inline explicit event_t(const context_t& ctx, int timeout = 0) noexcept :
    ctx_(*ctx), ptr_(std::make_shared<data>(*ctx, timeout)) {
        evt_ = ptr_->event;
    }

    inline event_t() noexcept :
    ptr_(nullptr) {}

    inline explicit operator const eXosip_event_t *() const { return evt_;}
    inline explicit operator bool() const { return evt_ != nullptr;}
    inline auto operator!() const { return evt_ == nullptr;}
    inline auto operator*() const { return evt_;}
    inline auto operator->() const { return evt_;}

    inline auto operator==(const event_t& rhs) {
        return evt_ == rhs.evt_;
    }

    inline auto operator!=(const event_t& rhs) {
        return evt_ != rhs.evt_;
    }

    inline auto context() const {
        return ctx_;
    }

    inline auto type() const {
        return evt_ ? evt_->type : -1;
    }

    inline auto text() const {
        return evt_ ? evt_->textinfo : "timeout";
    }

    inline auto request() const {
        return evt_ ? evt_->request : nullptr;
    }

    inline auto response() const {
        return evt_ ? evt_->response : nullptr;
    }

    inline auto ack() const {
        return evt_ ? evt_->ack : nullptr;
    }

    inline auto method() const {
        return request() ? evt_->request->sip_method : "NONE";
    }

    inline auto status() const {
        return response() ? evt_->response->status_code : 407;
    }

    inline auto timeout() const {
        return evt_ == nullptr;
    }

    inline auto tid() const {
        return evt_ ? evt_->tid : -1;
    }

    inline auto cid() const {
        return evt_ ? evt_->cid : -1;
    }

    inline auto did() const {
        return evt_ ? evt_->did : -1;
    }

    inline auto automatic() const {
        guard_t lock(ctx_);
        return eXosip_default_action(ctx_, evt_) == 0;
    }

protected:
    struct data {
        eXosip_event_t *event;

        inline explicit data(eXosip_t *ctx, int timeout) noexcept : event(eXosip_event_wait(ctx, timeout / 1000,  timeout % 1000)) {}

        inline ~data() {
            if(event)
                eXosip_event_free(event);
        }

        data(const data&) = delete;
        auto operator=(const data&) -> data& = delete;
    };

    eXosip_event_t *evt_{};
    eXosip_t *ctx_{};
    std::shared_ptr<data> ptr_;
};

} // end namespace

/*!
 * Namespace for eXosip2 stack.
 * \namespace exosip
 */

/*!
 * Header only utility c++ library for the eXosip2 stack.
 * \file exosip.hpp
 */
#endif
